import logging
from types import NoneType

import numpy as np
import torch
import os
from fitxf.math.fit.arc.ClassifierArcInterface import ClassifierArcInterface
from fitxf.math.utils.Logging import Logging


class ClassifierArcUnitTest:

    def __init__(
            self,
            child_class: type(ClassifierArcInterface),
            logger = None,
    ):
        self.child_class = child_class
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(
            self,
    ):
        accuracies = {}
        for f in [
            # 'max',
            'sum',
        ]:
            accuracies[f] = self.test_by_func(
                load_state_if_exists = False,
                test_function = f,
            )
        self.logger.info('Tests passed with accuracies ' + str(accuracies))
        return

    def test_by_func(
            self,
            load_state_if_exists = False,
            # max, sum
            test_function: str = 'max',
    ):
        X = torch.rand(size=(1024, 4))

        if test_function == 'max':
            is_categorical = True
            # category is just the largest index
            y, n_cat = torch.argmax(X, dim=-1), X.shape[-1]
            assert len(np.unique((y))) == n_cat
            # y_onehot_or_value = torch.nn.functional.one_hot(y, num_classes=n_cat).to(torch.float)
            n_hidden_features = (100, 50,)
            hidden_functions = (torch.nn.Linear, torch.nn.Linear, torch.nn.Linear)
            # for max function, it can be any function that is always increasing, thus we choose
            # Tanh() function since it is nicely bounded & satisfies always increasing
            activation_functions = (torch.nn.Tanh, torch.nn.Tanh, torch.nn.Softmax)
            loss_f = torch.nn.CrossEntropyLoss
            dropout = 0.2
            learn_rate = 0.001
            regularization_type = 0.
            num_epochs, batch_sz = 10, 16
            acc_thr = 0.90
        else:
            is_categorical = True
            # category is the sum of the rounded X
            y, n_cat = torch.sum(torch.round(X).to(torch.int), dim=-1), X.shape[-1] + 1
            assert len(np.unique((y))) == n_cat
            # y_onehot_or_value = torch.unsqueeze(y_onehot_or_value, dim=1)
            # y_onehot_or_value = torch.nn.functional.one_hot(y, num_classes=n_cat).to(torch.float)
            n_hidden_features = (100, None)
            hidden_functions = (torch.nn.Linear, None, torch.nn.Linear)
            # since summation is a linear function, any non-linear activation will cause problems
            activation_functions = (None, None, torch.nn.Softmax)
            loss_f = torch.nn.CrossEntropyLoss
            dropout = 0.
            learn_rate = 0.001
            regularization_type = 0.
            # for summation, need smaller batch sizes, otherwise can't converge
            num_epochs, batch_sz = 10, 4
            acc_thr = 0.60

        assert len(X) == len(y)
        self.logger.info(
            'X input (shape ' + str(X.shape) + '):\n' + str(X) + '\ny output (shape ' + str(y.shape)
            + '):\n' + str(y)
        )
        clf = self.child_class(
            in_features = X.shape[-1],
            out_features = n_cat,
            n_hidden_features = n_hidden_features,
            dropout_rate = dropout,
            learning_rate = learn_rate,
            hidden_functions = hidden_functions,
            activation_functions = activation_functions,
            loss_function = loss_f,
            logger = self.logger,
        )
        model_filepath = 'ut.' + str(self.child_class.__name__) + '.bin'

        if load_state_if_exists and os.path.exists(model_filepath):
            clf.from_old_states(
                model_filepath = model_filepath,
            )
            out_args, out_vals = clf.predict(X=X)
            # take 1st value of every row
            out_cat_top = out_args[:,:1].flatten()
            self.logger.info('Categories top predicted: ' + str(out_cat_top) + ', shape ' + str(out_cat_top.shape))
            self.logger.info('y: ' + str(y) + ', shape ' + str(y.shape))
            correct = 1 * (y - out_cat_top == 0)
            self.logger.info('Correct: ' + str(correct))
            acc = torch.sum(correct) / len(correct)
            self.logger.info(
                'Evaluation results for "' + str(self.child_class.__name__) + '": Total correct '
                + str(torch.sum(correct).item()) + ' from length ' + str(len(correct)) + ', accuracy ' + str(acc.item())
            )
        else:
            res = clf.fit(
                X = X,
                y = y,
                is_categorical = is_categorical,
                # we already calculated onehot ourselves
                num_categories = n_cat,
                batch_size = batch_sz,
                epochs = num_epochs,
                regularization_type = regularization_type,
            )
            acc = res['eval_accuracy']
        # [self.logger.info(p) for p in clf.parameters()]
        self.logger.info(
            'Child class "' + str(self.child_class.__name__) + '" accuracy ' + str(acc)
            + ' for test "' + str(test_function) + '"'
        )
        assert acc > acc_thr, \
            'Child class "' + str(self.child_class.__name__) + '" Accuracy from evaluation ' \
            + str(acc) + ' not > ' + str(acc_thr) + ' for test "' + str(test_function) + '"'

        if not os.path.exists(model_filepath):
            clf.save_states(
                model_filepath = model_filepath,
                additional_info = {},
            )
        self.logger.info('TEST PASSED FOR "' + str(test_function) + '"')
        return acc


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    exit(0)



# -*- coding: utf-8 -*-
import logging
import smtplib
import os
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from fitxf.math.utils.Log import Logging


class SendMail:

    COMMASPACE = ', '

    PORT_SSL = 465
    PORT_SMTP = 587
    GMAIL_SMTP = 'smtp.gmail.com'

    MAX_TOTAL_FILES_SIZE_MB_EMAIL_ATTCH = 25

    MAIL_MODE_SSL  = 'ssl'
    MAIL_MODE_SMTP = 'smtp'
    MAIL_MODE_SMTP_STARTTLS = 'smtp-starttls'

    def __init__(
            self,
            mode             = MAIL_MODE_SMTP,
            mail_server_url  = GMAIL_SMTP,
            mail_server_port = PORT_SMTP,
            logger = None,
    ):
        self.mode = mode
        self.mail_server_url = mail_server_url
        self.mail_server_port = mail_server_port
        self.logger = logger if logger is not None else logging.getLogger()
        self.__init_smtp()
        return

    def __init_smtp(self):
        self.logger.info(
            'Trying to initialize mail server "' + str(self.mail_server_url)
            + '" port ' + str(self.mail_server_port) + ' using mode "' + str(self.mode) + '"...'
        )
        if self.mode == self.MAIL_MODE_SSL:
            # Create a secure SSL context
            # self.context = ssl.create_default_context()
            self.server = smtplib.SMTP_SSL(
                host = self.mail_server_url,
                port = self.mail_server_port,
                # context=self.context
            )
            self.server.ehlo()
        elif self.mode == self.MAIL_MODE_SMTP:
            self.server = smtplib.SMTP(
                host=self.mail_server_url,
                port=self.mail_server_port
            )
            self.server.ehlo()
        else:
            self.server = smtplib.SMTP(
                host = self.mail_server_url,
                port = self.mail_server_port
            )
            self.server.ehlo()
            self.server.starttls()
        self.logger.info(
            'SMTP mode "' + str(self.mode) + '" successfully initialized.'
        )
        return

    def prepare_message(
            self,
            from_addr,
            to_addrs_list,
            subject,
            text,
            files = None,
    ):
        try:
            msg = MIMEMultipart()
            msg['From'] = from_addr
            msg['To'] = SendMail.COMMASPACE.join(to_addrs_list)
            msg['Date'] = formatdate(localtime=True)
            msg['Subject'] = subject

            if text[0]=='<':
                msg.attach(MIMEText(text, 'html'))
            else:
                msg.attach(MIMEText(text, 'plain'))

            files_allowed = self.__attach_file_check_validity_and_size(
                files_attachment_list = files,
                max_total_files_size = SendMail.MAX_TOTAL_FILES_SIZE_MB_EMAIL_ATTCH
            )

            for f in files_allowed or []:
                with open(f, "rb") as fil:
                    part = MIMEApplication(
                        fil.read(),
                        Name = os.path.basename(f)
                    )
                # After the file is closed
                part['Content-Disposition'] = 'attachment; filename="%s"' % os.path.basename(f)
                msg.attach(part)
            return msg.as_string()
        except Exception as ex:
            errmsg = 'Error creating email message: ' + str(ex)
            self.logger.error(errmsg)
            raise Exception(errmsg)
        #message = """From: %s\nTo: %s\nSubject: %s\n\n%s
        #    """ % (from_addr, ", ".join(to_addrs_list), subject, text)
        #return message

    def __attach_file_check_validity_and_size(
            self,
            files_attachment_list,
            max_total_files_size = MAX_TOTAL_FILES_SIZE_MB_EMAIL_ATTCH
    ):
        if files_attachment_list is None:
            return []

        files_attachment_list_allowed = []

        cum_size_mb = 0.0
        for filepath in files_attachment_list:
            if os.path.isfile(filepath):
                self.logger.info(
                    'Attachment file path "' + str(filepath) + '" OK'
                )
            else:
                self.logger.error(
                    'Invalid attachment file "' + str(filepath) + '", not attaching to email'
                )
                continue

            fsize_bytes = os.path.getsize(filepath)
            fsize_mb = round(fsize_bytes / (1024 * 1024), 2)

            if fsize_mb+cum_size_mb < max_total_files_size:
                files_attachment_list_allowed.append(filepath)
                cum_size_mb += fsize_mb
                self.logger.info(
                    'Appended file "' + str(filepath) + '" as email attachment size ' + str(fsize_mb)
                    + 'MB, total cumulative ' + str(cum_size_mb) + 'MB'
                )
            else:
                self.logger.warning(
                    'File "' + str(filepath) + '" too big ' + str(fsize_mb)
                    + 'MB. Cumulative = ' + str(fsize_mb+cum_size_mb) + ' Not attaching to email'
                )
        return files_attachment_list_allowed

    def send(
            self,
            user,
            password,
            recipients_list,
            message
    ):
        try:
            if password not in [None, '']:
                self.server.login(
                    user = user,
                    password = password
                )
                self.logger.info(
                    'Login for user "' + str(user) + '" successful.'
                )
            else:
                # If no password passed in, no need to do login
                self.logger.warning(
                    'Not doing login for user "' + str(user) + '", no password given "' + str(password) + '"'
                )
            self.server.sendmail(
                from_addr = user,
                to_addrs  = recipients_list,
                msg       = message
            )
            self.logger.info(
                'Message from '+ str(user) + ' to ' + str(recipients_list)
                + ' sent successfully. Closing server..'
            )
            self.server.close()
            self.logger.info(
                'Mail server "' + str(self.mail_server_url) + '" closed'
            )
        except Exception as ex:
            errmsg = 'Exception sending mail from ' + str(user) + ' to ' + str(recipients_list)\
                     + '. Got exception ' + str(ex) + '.'
            self.logger.error(errmsg)
            raise Exception(errmsg)


if __name__ == '__main__':
    user = 'mapktah.naver.com@gmail.com'
    receivers = ['mapktah@mail.ru']
    subject = 'Test mail Python'
    text = 'Test message from Python client (CAI auto alerts)'
    sm = SendMail(
        mode = SendMail.MAIL_MODE_SMTP_STARTTLS,
        mail_server_url = SendMail.GMAIL_SMTP,
        mail_server_port = SendMail.PORT_SMTP,
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False),
    )
    message = sm.prepare_message(
        from_addr = user,
        to_addrs_list = receivers,
        subject = subject,
        text = text,
        files = []
    )

    # message = """From: From Kim Bon <kimbon@gmail.com>
    # To: To All
    # Subject: SMTP e-mail test
    #
    # This is a test e-mail message.
    # """

    sm.send(
        user = user,
        password = 'fshg fvzo aelw xath',
        recipients_list = receivers,
        message = message
    )
    exit(0)



import math
import logging
import torch
import numpy as np
from fitxf.math.fit.utils.FitUtils import FitUtils
from fitxf.math.utils.Logging import Logging
from torch.utils.data import DataLoader, RandomSampler


class TestModel(torch.nn.Module):

    def __init__(self):
        super().__init__()
        self.h1 = torch.nn.Linear(in_features=3, out_features=100)
        self.h2 = torch.nn.ReLU()
        self.h3 = torch.nn.Linear(in_features=100, out_features=1)
        return

    def forward(self, X):
        y1 = self.h1(X)
        y2 = self.h2(y1)
        y3 = self.h3(y2)
        return y3


class FitUtilsUt:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def _test_pred_helper(
            self,
            test_name,
            X_test,
            Y_test,
            # list of tuples (4 elements: normalize_or_not, cluster_centers, predicted_classes, predicted_distances)
            test_set,
            mean_per_label,
            remove_dup_lbl,
    ):
        # if not test_name in ['Test 2 Mean Per Label #1']:
        #     return
        fut = FitUtils(logger=self.logger)
        for (nm, exp_val, exp_cls, exp_dist) in test_set:
            mean_ts = fut.fit_mean(
                X=X_test, labels=Y_test, mean_per_label=mean_per_label, normalize_X=nm
            )
            mean_ts_list = torch.round(mean_ts, decimals=5).tolist()
            exp_val_list = torch.round(exp_val, decimals=5).tolist()
            # Test by class
            for class_id in torch.unique(Y_test):
                # index must be integral
                class_id = int(class_id.item())
                # convert to list and sort for consistency, because the clusters may return in non-specific ordering
                mean_tx_class_id = mean_ts_list[class_id]
                mean_tx_class_id.sort()
                exp_val_class_id = exp_val_list[class_id]
                exp_val_class_id.sort()
                print(
                    'Checking class no. ' + str(class_id) + ' Fitted mean per label = ' + str(mean_per_label)
                    +  ', normalize = ' + str(nm) + ' with X:\n' + str(X_test[Y_test==class_id])
                    + '\nresult fit:\n' + str(mean_tx_class_id) + '\nexpected:\n' + str(exp_val_class_id)
                )
                # element wise difference, all must be exactly 0.
                test_tsr = torch.round(torch.Tensor(mean_tx_class_id), decimals=4) - torch.Tensor(exp_val_class_id)
                test_res = torch.sum(test_tsr != 0.) == 0
                assert test_res, \
                    ('<' + str(test_name) + '> For class no. ' + str(class_id) + ', normalize=' + str(nm)
                     + ', expect mean ' + str(exp_val_class_id) + ' observed ' + str(mean_tx_class_id))

            test_top_k = 3
            c, d = fut.predict_mean(
                X              = X_test,
                mean_tensors   = mean_ts,
                reps_per_label = mean_per_label,
                remove_dup_lbl = remove_dup_lbl,
                normalize_X    = nm,
            )
            # print('***** Predicted classes')
            # print(c)
            if exp_cls is not None:
                for i, c_row in enumerate(c):
                    # Test top k classes will do
                    if i >= test_top_k:
                        break
                    test_tsr = torch.sum((c_row - exp_cls[i]) != 0.)
                    test_res = test_tsr == 0.
                    assert test_res, \
                        '<' + str(test_name) + '> Row #' + str(i) + ' For normalize=' \
                        + str(nm) + ', test tensor:\n' + str(test_tsr) \
                        + '\n, expect classes:\n' + str(exp_cls[i]) + '\nobserved:\n' + str(c_row)
            if exp_dist is not None:
                for i, d_row in enumerate(d):
                    # Test top k classes will do
                    if i >= test_top_k:
                        break
                    test_tsr = torch.sum((torch.round(d_row, decimals=4) - exp_dist[i]) != 0.)
                    test_res = test_tsr == 0.
                    assert test_res, \
                        '<' + str(test_name) + '> Row #' + str(i) + ' For normalize=' + str(nm) \
                        + ', test tensor:\n' + str(test_tsr) + '\n, expect dist:\n' + str(exp_dist[i]) \
                        + '\nobserved:\n' + str(d_row)

    def test_predict(
            self,
    ):
        _, _, lbl_mapped = FitUtils().map_labels_to_consecutive_numbers(lbl_list=[4, 2, 6, 6, 6, 55, 0])
        expected_res = [2, 1, 3, 3, 3, 4, 0]
        assert lbl_mapped == expected_res, \
            'Test map labels to consecutive numbers, observed ' + str(lbl_mapped) + ' expected ' + str(expected_res)

        X = torch.Tensor([
            # class 0
            [0, 1], [1, 2], [2, 3], [4, 5], [5, 6],
            # class 1
            [6, 7], [7, 8], [8, 9],
            # class 2
            [9, 10], [10, 11],
        ])
        # at_msk = torch.Tensor([True] * 10)
        Y = torch.Tensor([0, 0, 0, 0, 0, 1, 1, 1, 2, 2])
        # class mean tensors after classification
        expected_value = torch.Tensor([
            [2.4000, 3.4000],
            [7.0000, 8.0000],
            [9.5000, 10.5000],
        ])
        # normalized class mean tensors
        expected_value_nm = torch.Tensor([
            [0.4534, 0.8551],
            [0.6579, 0.7531],
            [0.6708, 0.7416],
        ])
        expected_classes = torch.IntTensor([
            [0., 1., 2.],
            [0., 1., 2.],
            [0., 1., 2.],
            [0., 1., 2.],
            # with only 1 mean per label, the 5th point is predicted wrong
            [1., 0., 2.],
            [1., 2., 0.],
            [1., 2., 0.],
            [1., 2., 0.],
            [2., 1., 0.],
            [2., 1., 0.]
        ])
        # normalizing will screw up predictions
        expected_classes_nm = torch.IntTensor([
            [0, 1, 2],
            [1, 0, 2],
            [1, 2, 0],
            [1, 2, 0],
            [1, 2, 0],
            [1, 2, 0],
            [1, 2, 0],
            [2, 1, 0],
            [2, 1, 0],
            [2, 1, 0]
        ])
        # from euclidean distance
        expected_distances = torch.Tensor([
            [3.3941, 9.8995, 13.4350],
            [1.9799, 8.4853, 12.0208],
            [0.5657, 7.0711, 10.6066],
            [2.2627, 4.2426, 7.7782],
            [2.8284, 3.6770, 6.3640],
            [1.4142, 4.9497, 5.0912],
            [0.0000, 3.5355, 6.5054],
            [1.4142, 2.1213, 7.9196],
            [0.7071, 2.8284, 9.3338],
            [0.7071, 4.2426, 10.7480]
        ])
        # from dot product
        expected_distances_nm = torch.Tensor([
            [0.8551, 0.7531, 0.7416],
            [0.9678, 0.9676, 0.9633],
            [0.9915, 0.9892, 0.9630],
            [0.9990, 0.9982, 0.9509],
            [0.9997, 0.9992, 0.9471],
            [0.9999, 0.9996, 0.9443],
            [1.0000, 0.9999, 0.9421],
            [1.0000, 0.9999, 0.9403],
            [1.0000, 0.9999, 0.9389],
            [1.0000, 0.9998, 0.9377]
        ])
        self._test_pred_helper(
            test_name = 'Test 1 Mean Per Label',
            X_test   = X,
            Y_test   = Y,
            test_set = [
                (False, expected_value, expected_classes, expected_distances),
                (True, expected_value_nm, expected_classes_nm, expected_distances_nm),
            ],
            mean_per_label = 1,
            remove_dup_lbl = False,
        )

        expected_value2 = torch.Tensor([
            [[4.5000, 5.5000], [1.0000, 2.0000]],
            [[6.5000, 7.5000], [8.0000, 9.0000]],
            [[9., 10.], [10., 11.]],
        ])
        expected_value2_nm = torch.Tensor([
            [[0.5667, 0.8189], [0.0000, 1.0000]],
            [[0.6508, 0.7593], [0.6614, 0.7500]],
            [[0.6690, 0.7433], [0.6727, 0.7399]],
        ])
        expected_classes2 = torch.IntTensor([
            [0, 0, 1, 1, 2],
            [0, 0, 1, 1, 2],
            [0, 0, 1, 1, 2],
            [0, 1, 0, 1, 2],
            [0, 1, 1, 0, 2],
            [1, 0, 1, 2, 2],
            [1, 1, 2, 0, 2],
            [1, 2, 1, 2, 0],
            [2, 1, 2, 1, 0],
            [2, 2, 1, 1, 0]
        ])
        expected_classes_rmdup2 = torch.IntTensor([
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [1, 0, 2],
            [1, 2, 0],
            [1, 2, 0],
            [2, 1, 0],
            [2, 1, 0]
        ])
        self._test_pred_helper(
            test_name = 'Test 2 Mean Per Label #1',
            X_test   = X,
            Y_test   = Y,
            test_set = [
                (False, expected_value2, expected_classes2, None),
                (True, expected_value2_nm, None, None),
            ],
            mean_per_label = 2,
            remove_dup_lbl = False,
        )
        # with remove duplicate predictions
        self._test_pred_helper(
            test_name = 'Test 2 Mean Per Label #2',
            X_test   = X,
            Y_test   = Y,
            test_set = [
                (False, expected_value2, expected_classes_rmdup2, None),
                (True, expected_value2_nm, None, None),
            ],
            mean_per_label = 2,
            remove_dup_lbl = True,
        )

        #
        # Test for cluster centers more than points
        #
        expected_value3 = torch.Tensor([
            [[0.5000, 1.5000], [4.5000, 5.5000], [2.0000, 3.0000]],
            [[6.0000, 7.0000], [8.0000, 9.0000], [7.0000, 8.0000]],
            [[9.0, 10.0],      [9.0, 10.0],      [10.0, 11.0]],
        ])
        expected_value3_nm = torch.Tensor([
            [[0.6065, 0.7937], [0.0000, 1.0000], [0.4472, 0.8944]],
            [[0.6508, 0.7593], [0.6644, 0.7474], [0.6585, 0.7526]],
            [[0.6690, 0.7433], [0.6690, 0.7433], [0.6727, 0.7399]],
        ])
        # when mean tensors not normalized & prediction using euclid-dist, all predictions correct
        expected_classes3 = torch.IntTensor([
            [0, 0, 0, 1, 1],
            [0, 0, 0, 1, 1],
            [0, 0, 0, 1, 1],
            [0, 0, 1, 1, 0],
            [0, 1, 1, 1, 0],
            [1, 1, 0, 1, 2],
            [1, 1, 1, 2, 0],
            [1, 2, 1, 1, 2],
            [2, 2, 2, 1, 1],
            [2, 2, 2, 1, 1]
        ])
        expected_classes_rmdup3 = torch.IntTensor([
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [1, 0, 2],
            [1, 2, 0],
            [1, 2, 0],
            [2, 1, 0],
            [2, 1, 0]
        ])
        expected_classes_rmdup3_nm = torch.IntTensor([
            [0, 1, 2],
            [0, 1, 2],
            [0, 1, 2],
            [1, 0, 2],    # wrong pred
            [1, 2, 0],    # wrong pred
            [1, 2, 0],
            [1, 2, 0],
            [1, 2, 0],
            [2, 1, 0],
            [2, 1, 0]
        ])
        expected_classes3_nm = torch.IntTensor([
            [0, 0, 0, 1, 1],
            [0, 0, 1, 1, 1],
            [0, 1, 0, 1, 1],
            [1, 1, 0, 1, 2],    # wrong pred
            [1, 1, 1, 2, 2],    # wrong pred
            [1, 1, 1, 2, 2],
            [1, 1, 1, 2, 2],
            [1, 2, 1, 2, 2],
            [2, 2, 2, 1, 1],
            [2, 2, 2, 1, 1]
        ])
        expected_distances3 = torch.Tensor([
            [0.7071, 2.8284, 6.3640, 8.4853, 9.8995],
            [0.7071, 1.4142, 4.9497, 7.0711, 8.4853],
            [0.0000, 2.1213, 3.5355, 5.6569, 7.0711],
            [0.7071, 2.8284, 2.8284, 4.2426, 4.9497],
            [0.7071, 1.4142, 2.8284, 4.2426, 4.2426],
            [0.0000, 1.4142, 2.1213, 2.8284, 4.2426],
            [0.0000, 1.4142, 1.4142, 2.8284, 3.5355],
            [0.0000, 1.4142, 1.4142, 2.8284, 2.8284],
            [0.0000, 1.4142, 1.4142, 1.4142, 2.8284],
            [0.0000, 0.0000, 1.4142, 2.8284, 4.2426]
        ])
        expected_distances_rmdup3 = torch.Tensor([
            [0.7071, 8.4853, 12.7279],
            [0.7071, 7.0711, 11.3137],
            [0.0000, 5.6569, 9.8995],
            [0.7071, 2.8284, 7.0711],
            [0.7071, 1.4142, 5.6569],
            [0.0000, 2.1213, 4.2426],
            [0.0000, 2.8284, 3.5355],
            [0.0000, 1.4142, 4.9497],
            [0.0000, 1.4142, 6.3640],
            [0.0000, 2.8284, 7.7782]
        ])
        # from dot product
        expected_distances3_nm = torch.Tensor([
            [1.0000, 0.8944, 0.7937, 0.7593, 0.7526],
            [1.0000, 0.9812, 0.9701, 0.9676, 0.9656],
            [0.9968, 0.9927, 0.9923, 0.9915, 0.9904],
            [0.9994, 0.9990, 0.9987, 0.9987, 0.9983],
            [0.9999, 0.9997, 0.9995, 0.9993, 0.9991],
            [1.0000, 0.9999, 0.9998, 0.9997, 0.9996],
            [1.0000, 1.0000, 0.9999, 0.9999, 0.9998],
            [1.0000, 1.0000, 1.0000, 0.9999, 0.9999],
            [1.0000, 1.0000, 1.0000, 1.0000, 0.9999],
            [1.0000, 1.0000, 1.0000, 0.9999, 0.9998]
        ])
        expected_distances_rmdup3_nm = torch.Tensor([
            [1.0000, 0.7593, 0.7433],
            [1.0000, 0.9701, 0.9640],
            [0.9968, 0.9927, 0.9895],
            [0.9994, 0.9987, 0.9983],
            [0.9999, 0.9993, 0.9980],
            [1.0000, 0.9997, 0.9974],
            [1.0000, 0.9999, 0.9967],
            [1.0000, 1.0000, 0.9962],
            [1.0000, 1.0000, 0.9957],
            [1.0000, 0.9999, 0.9953]
        ])
        self._test_pred_helper(
            test_name = 'Test 3 Mean Per Label (remove dup lbl = False)',
            X_test   = X,
            Y_test   = Y,
            test_set = [
                (False, expected_value3, expected_classes3, expected_distances3),
                (True, expected_value3_nm, expected_classes3_nm, expected_distances3_nm),
            ],
            mean_per_label = 3,
            remove_dup_lbl = False,
        )
        self._test_pred_helper(
            test_name = 'Test 3 Mean Per Label (remove dup lbl = True)',
            X_test   = X,
            Y_test   = Y,
            test_set = [
                (False, expected_value3, expected_classes_rmdup3, expected_distances_rmdup3),
                (True, expected_value3_nm, expected_classes_rmdup3_nm, expected_distances_rmdup3_nm),
            ],
            mean_per_label = 3,
            remove_dup_lbl = True,
        )

    def test_nn(self, epochs=5000, plot_graph=False, tol_dif=0.1):
        fu = FitUtils(logger=self.logger)
        """
        Sample train, train to recognize binary representation
        """
        X = np.array([
            [0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1],
        ], dtype=float)
        Y = np.array([
            [0], [1], [2], [3], [4], [5], [6], [7],
        ], dtype=float)
        # attn_masks = np.array([
        #     [1,1,1], [1,1,1], [1,1,1], [1,1,1], [1,1,1], [1,1,1], [1,1,1], [1,1,1],
        # ], dtype=float)

        X = torch.from_numpy(X).float()
        Y = torch.from_numpy(Y).float()
        # attn_masks = torch.from_numpy(attn_masks).float()

        train_set, val_set = fu.training_data_split(
            val_ratio  = 0.0,
            inputs     = X,
            labels     = Y,
            attn_masks = None,
            return_tensors = 'pt',
        )

        # Prepare DataLoader
        # Total step/batch should be int(n_train_samples/batch_size)
        dl = DataLoader(
            train_set,
            sampler = RandomSampler(train_set),
            batch_size = len(X),
        )

        for acc_steps in (1, len(X)/2,):
            acc_steps = math.floor(acc_steps)

            model = TestModel()
            model.train(mode=True)

            losses = fu.torch_train(
                train_dataloader = dl,
                model = model,
                loss_func = torch.nn.MSELoss(),
                optimizer = torch.optim.SGD(model.parameters(), lr=0.005),
                regularization_type = None,
                epochs = epochs,
                accum_steps = acc_steps,
                plot_losses = plot_graph,
            )
            # Final loss must not exceed difference of 0.01 per data
            tol = 7*(0.1**2)
            assert losses[-1] < tol, 'Final loss ' + str(losses[-1]) + ' must be less than ' + str(tol)
            print(
                'Accum steps ' + str(acc_steps) + ', initial loss ' + str(losses[0])
                + ' to final loss ' + str(losses[-1]) + ', tolerance ' + str(tol)
            )

            model.eval()
            pred = model(X)
            print('Expected/Predicted ' + str(list(zip(Y.tolist(), torch.round(pred, decimals=3).tolist()))))
            for i in range(len(Y)):
                x = X[i]
                val = Y[i][0]
                val_pred = pred[i][0]
                msg = 'For accum steps ' + str(acc_steps) + ', x=' + str(x) + ', expected ' + str(val)\
                      + ', got ' + str(val_pred) + ', tol=' + str(tol_dif)
                print(msg)
                assert abs(val_pred - val) < tol_dif, msg

    def test_dist(self):
        fu = FitUtils(logger=self.logger)
        x = np.array([
            [0., 0.], [0., 1.], [1., 0.],
        ])
        x_ref = np.array([0.2, 0.5])
        for ref, method, metric, exp_mean, exp_dists in [
            (None,  'exact', 'euclid',        1.138071187, np.array([1., 1., 1.41421356])),
            (None,  'exact', 'euclid-square', 4/3,         np.array([1., 1., 2.])),
            (x_ref, 'exact', 'euclid',        0.673477024, np.array([0.53851648, 0.53851648, 0.94339811])),
            (x_ref, 'exact', 'euclid-square', 0.49,         np.array([0.29, 0.29, 0.89])),
            (None,  'mc',    'euclid',        1.138071187, None),
            (None,  'mc',    'euclid-square', 4/3,         None),
            (x_ref, 'mc',    'euclid',        0.673477024, None),
            (x_ref, 'mc',    'euclid-square', 0.49,        None),
        ]:
            dists = fu.get_point_distances(
                np_tensors = x,
                np_center  = ref,
                method     = method,
                metric     = metric,
                mc_iters   = 10000,
            )
            dists_mean = np.mean(dists)
            print('Method', method, 'metric', metric, dists[0:min(30,len(dists))])
            print('Mean', dists_mean)
            if exp_dists is not None:
                assert np.sum(np.round(exp_dists - dists, 4)**2) == 0., \
                    'Distances method "' + str(method) + '", metric "' + str(metric) + '":' + str(dists)
            assert round((exp_mean - dists_mean)**2, 2) == 0., \
                'Mean distance method "' + str(method) + '", metric "' + str(metric) + '":' + str(dists_mean)

    def test_map_to_connbr(self):
        fu = FitUtils(logger=self.logger)
        labels = ['drama', 'comedy', 'action', 'comedy', 'action', 'drama']
        test_sets = [
            {
                'curmap': None,
                'expected_map': {'action': 0, 'comedy': 1, 'drama': 2},
            },
            {
                'curmap': {'comedy': 1},
                'expected_map': {'comedy': 1, 'action': 0, 'drama': 2},
            },
            {
                'curmap': {'comedy': 99},
                'expected_map': {'comedy': 99, 'action': 0, 'drama': 1},
            },
            {
                'curmap': {'horror': 99},
                'expected_map': {'action': 0, 'comedy': 1, 'drama': 2, 'horror': 99},
            },
            {
                'curmap': {'action': 77, 'comedy': 88, 'drama': 99},
                'expected_map': {'action': 77, 'comedy': 88, 'drama': 99},
            },
            {
                'curmap': {'comedy': 88, 'drama': 99, 'horror': 111},
                'expected_map': {'action': 0, 'comedy': 88, 'drama': 99, 'horror': 111},
            },
            {
                # Should throw exception with duplicate values
                'curmap': {'comedy': 0, 'drama': 0, 'horror': 1},
                'expected_map': None,
            },
        ]
        for i, s in enumerate(test_sets):
            curmap, exp_map_lbl = s['curmap'], s['expected_map']
            try:
                map_lbl = fu.map_labels_to_numbers(lbl_list=labels, existing_label_map=curmap, strict_1_to_1=True)
            except:
                map_lbl = None
            assert map_lbl == exp_map_lbl, \
                '# ' + str(i) + ' Observed map ' + str(map_lbl) + ' different from expected ' + str(exp_map_lbl)

    def test(self):
        self.test_map_to_connbr()
        self.test_nn(epochs=5000, plot_graph=False, tol_dif=0.1)
        self.test_predict()
        self.test_dist()
        print('ALL TESTS PASSED')


if __name__ == '__main__':
    fu = FitUtilsUt(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False))
    fu.test()
    exit(0)



import logging
import matplotlib.pyplot as plt
import numpy as np
from scipy import signal
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


class DspFilter:

    def __init__(
            self,
            logger: Logging = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def butter_bandpass(self, lowcut, highcut, fs, order=5):
        b, a = signal.butter(order, [lowcut, highcut], btype='band', fs=fs)
        return b, a

    def butter_bandpass_filter(self, data: np.ndarray, lowcut, highcut, fs, order=5):
        b, a = self.butter_bandpass(lowcut, highcut, fs, order=order)
        y = signal.lfilter(b, a, data)
        return y

    def bandpass_filter(self, buffer: np.ndarray, lowcut, highcut, fs):
        return self.butter_bandpass_filter(data=buffer, lowcut=lowcut, highcut=highcut, fs=fs, order=6)


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env')
    exit(0)



import logging
import os
from fitxf.math.datasource.DatastoreInterface import DbParams, DatastoreInterface


class ModelDbInterface:

    def __init__(
            self,
            tablename: str,
            col_content: str,
            col_content_type: str,
            col_label_user: str,
            col_label_standardized: str,
            col_embedding: str,
            max_records: int,
            logger = None,
    ):
        self.tablename = tablename
        self.col_content = col_content
        self.col_content_type = col_content_type
        self.col_label_user = col_label_user
        self.col_label_standardized = col_label_standardized
        self.col_embedding = col_embedding
        self.max_records = max_records
        self.logger = logger if logger is not None else logging.getLogger()

        create_tbl_sql = self.get_create_table_db_cmd(db_type=os.environ["DB_TYPE"])
        self.db_params = DbParams.get_db_params_from_envvars(
            identifier = tablename,
            db_create_tbl_sql = create_tbl_sql,
            db_table = tablename,
            verify_certs = os.environ.get("VERIFY_CERTS", "1") in ('1', 'yes', 'true'),
        )

        return

    def get_create_table_db_cmd(
            self,
            db_type,
    ) -> str:
        raise Exception('물려받는 클래스에서 구현해야 함')

    def get_db_params(self) -> DbParams:
        return self.db_params

    def get_underlying_db(self) -> DatastoreInterface:
        raise Exception('물려받는 클래스에서 구현해야 함')

    def insert(
            self,
            records: list[dict],
            tablename: str,
    ):
        raise Exception('물려받는 클래스에서 구현해야 함')

    def delete(
            self,
            match_phrase: dict,
            tablename: str,
    ):
        raise Exception('물려받는 클래스에서 구현해야 함')

    def load_data(
            self,
            max_attempts = 1,
    ):
        raise Exception('물려받는 클래스에서 구현해야 함')

    def connect_to_underlying_db(
            self,
    ):
        raise Exception('물려받는 클래스에서 구현해야 함')

    def convert_csv_string_array_to_float_array(
            self,
            string_array,
            custom_chars_remove,
    ):
        raise Exception('물려받는 클래스에서 구현해야 함')


if __name__ == '__main__':
    exit(0)



import logging
import numpy as np
import torch
import joblib
from fitxf.math.fit.arc.ClassifierArcInterface import ClassifierArcInterface
from fitxf.math.fit.arc.ClassifierArcUnitTest import ClassifierArcUnitTest
from sklearn.ensemble import RandomForestClassifier
from fitxf.math.utils.Logging import Logging


class ClassifierArcRf(ClassifierArcInterface):

    USE_ARRAY = False

    def __init__(
            self,
            model_filepath: str = None,
            in_features: int = None,
            out_features: int = None,
            n_hidden_features: int = 100,
            hidden_functions: list = (torch.nn.Linear, torch.nn.Linear, torch.nn.Linear),
            activation_functions: list = (torch.nn.ReLU, torch.nn.ReLU, torch.nn.Softmax),
            loss_function = torch.nn.CrossEntropyLoss,
            dropout_rate: float = 0.2,
            learning_rate: float = 0.0001,
            logger = None,
    ):
        super().__init__(
            model_filepath = model_filepath,
            logger = logger,
        )
        self.model_rf = None
        if self.model_filepath is not None:
            self.from_old_states(model_filepath=self.model_filepath)
        return

    def from_old_states(
            self,
            model_filepath: str,
    ):
        self.model_rf = joblib.load(model_filepath)
        self.logger.info('Loaded old state for random forest model from file "' + str(self.model_filepath) + '"')
        return self.model_rf

    def save_states(
            self,
            model_filepath: str,
            additional_info: dict,
    ):
        joblib.dump(self.model_rf, model_filepath)
        self.logger.info('Saved random forest model to "' + str(model_filepath) + '"')
        return

    def forward(
            self,
            x: torch.Tensor,
    ):
        return self.predict(X=x)

    def fit(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            is_categorical: bool = True,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
            # the smaller the batch size, the smaller the losses will be during training
            batch_size: int = 32,
            epochs: int = 100,
            eval_percent: float = 0.2,
            # important to prevent over-fitting
            regularization_type = "L2",
    ):
        n_cutoff_x_train = int(np.round((1 - eval_percent) * len(X)))
        self.logger.info(
            'X train cutoff = ' + str(n_cutoff_x_train) + ' total length = ' + str(len(X))
        )
        self.model_rf = RandomForestClassifier().fit(X[:n_cutoff_x_train], y[:n_cutoff_x_train])
        trees = self.model_rf.estimators_
        self.logger.info('Classified via random forest classifier, decision trees total ' + str(len(trees)))
        [self.logger.info(str('Tree #' + str(i) + ':' + str(dt))) for i, dt in enumerate(trees)]

        # Validate back
        out_cat, out_prob = self.predict(X[n_cutoff_x_train:])
        out_cat = out_cat.flatten()
        self.logger.info('Out cat shape ' + str(out_cat.shape) + ', y shape ' + str(y.shape))
        correct = 1 * (y[n_cutoff_x_train:] - out_cat == 0)
        eval_accuracy = torch.sum(correct) / len(correct)
        self.logger.info(
            'Evaluation results for random forest: Total correct ' + str(torch.sum(correct).item())
            + ' from length ' + str(len(correct)) + ', accuracy ' + str(eval_accuracy.item())
        )
        return {
            'eval_accuracy': eval_accuracy.item(),
            'losses': None,
            'dataloader_train': list(zip(X[:n_cutoff_x_train], y[:n_cutoff_x_train])),
            'dataloader_eval': list(zip(X[n_cutoff_x_train:], y[n_cutoff_x_train:])),
            'index_cutoff_train': n_cutoff_x_train,
        }

    def predict(
            self,
            X: torch.Tensor,
    ):
        # RF predicts only 1 category per row
        out_cat = self.model_rf.predict(X)
        out_cat_torch = torch.Tensor([[v] for v in out_cat])
        # self.logger.debug(out_cat_torch)
        out_prob_torch = torch.Tensor([[1.] for v in out_cat])
        return out_cat_torch, out_prob_torch


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    ut = ClassifierArcUnitTest(child_class=ClassifierArcRf, logger=lgr)
    ut.test()
    exit(0)



import os
import re
import logging
from fitxf.math.utils.File import FileUtils
from inspect import getsourcefile


class Env:

    @staticmethod
    def get_source_code_path():
        path = os.path.abspath(getsourcefile(lambda: 0))
        return path

    @staticmethod
    def get_home_dir():
        return os.path.expanduser("~")

    @staticmethod
    def get_home_download_dir():
        dir = str(Env.get_home_dir()) + os.sep + 'Downloads'
        if not os.path.isdir(dir):
            os.mkdir(dir)
        return dir

    @staticmethod
    def set_env_vars_from_file(env_filepath):
        assert os.path.isfile(env_filepath), 'Not a file "' + str(env_filepath) + '"'
        env_lines = FileUtils(filepath=env_filepath).read_text_file()
        env_lines_cleaned = [line for line in env_lines if line.strip()]
        env_lines_cleaned = [line for line in env_lines_cleaned if not re.match(pattern="^#", string=line)]
        print('Set environment variables from file "' + str(env_filepath) + '"')
        for line in env_lines_cleaned:
            varname, value = line.split(sep="=", maxsplit=1)
            os.environ[varname] = value
            print('Set env var ' + str(varname) + ' = "' + str(value) + '"')
        return

    def __init__(
            self,
            model_version = 'latest',
            logger = None,
    ):
        self.model_version = model_version
        self.logger = logger if logger is not None else logging.getLogger()

        try:
            import google.colab
            self.in_google_colab = True
            self.logger.info('Detected Colab environment')
        except:
            self.in_google_colab = False

        self.logger.info('In Google Colab "' + str(self.in_google_colab) + '"')

        if self.in_google_colab:
            self.REPO_DIR = '/content/drive/My Drive/colab/poc'
        else:
            self.REPO_DIR = self.guess_repo_dir(use_cwd_as_ref=True)
            self.logger.info('Not in any special environment, using repo dir "' + str(self.REPO_DIR) + '"')

        self.logger.info('Set to different environment, REPO_DIR "' + str(self.REPO_DIR))

        self.MODELS_PRETRAINED_DIR = self.REPO_DIR + '/_models/' + self.model_version
        self.CONFIG_DIR = self.REPO_DIR + r'/config'

        # ----- NLP DATASETS -----
        self.NLP_DATASET_DIR = self.REPO_DIR + r'/data/nlp-datasets'

    def guess_repo_dir(
            self,
            # Use current working directory as reference is more accurate since external non-related
            # projects can use this code and guess their directories correctly.
            # Otherwise if use source code, will only be correct for internal projects.
            use_cwd_as_ref = True,
    ):
        try:
            repo_dir = os.environ['REPO_DIR']
        except Exception as ex:
            self.logger.info('Failed to get repo directory from env var "REPO_DIR", got exception ' + str(ex))
            src_code_path = os.getcwd() if use_cwd_as_ref else self.get_source_code_path()
            self.logger.info('Try to guess repo dir from cwd "' + str(src_code_path) + '"')
            # Look "/src/" in Linux or "\src\" in Windows
            repo_dir = re.sub(pattern="(/src/.*)|([\\\\]src[\\\\].*)", repl="", string=src_code_path)
            print('Repository directory guessed as "' + str(repo_dir) + '"')
        return repo_dir


if __name__ == '__main__':
    print(Env.get_source_code_path())
    print(Env.get_home_dir())
    print(Env.get_home_download_dir())
    print(Env().REPO_DIR)
    exit(0)



import logging
import re
from fitxf.math.datasource.DatastoreInterface import DbParams
from fitxf.math.datasource.Csv import Csv
try:
    from fitxf.math.datasource.MySql import MySql
except Exception as __ex_import:
    print('Error importing library: ' + str(__ex_import))
from fitxf.math.utils.Env import Env


class Datastore:

    def __init__(
            self,
            db_params: DbParams,
            logger = None,
    ):
        self.db_params = db_params
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def get_data_store(
            self,
    ):
        if self.db_params.db_type == 'csv':
            assert not re.match(pattern="/", string=self.db_params.db_table), \
                'Must not contain full path in table name or index "' + str(self.db_params.db_table) + '"'
            DbClass = Csv
        elif self.db_params.db_type == 'mysql':
            DbClass = MySql
        else:
            raise Exception('Not supported data store type "' + str(self.db_params.db_type) + '"')

        self.logger.info(
            'Try to instantiate DB class with db params ' + str(self.db_params.get_db_info()) + ''
        )
        return DbClass(
            db_params = self.db_params,
            logger = self.logger,
        )


if __name__ == '__main__':
    Env.set_env_vars_from_file(env_filepath=Env().REPO_DIR + '/.env.fitxf.math.ut')
    dbp = DbParams.get_db_params_from_envvars(identifier='test', db_create_tbl_sql='', db_table='test_table')
    db = Datastore(db_params=dbp)
    db.get_data_store()
    exit(0)



import logging
import pickle
import os
import threading
import random
import time
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.LockF import LockFile


#
# Use cases:
#   1. RAM scalability
#      If you need to cache a 500M object in memory, you can use this class which will cache it in file
#      and quickly retrieve it when needed only.
#   2. Shared resource between multi-worker/threaded processes running on same server
#      If you have multiple workers and threads that need access to a shared object, this is it.
#
class ObjectPersistence:

    ATOMIC_UPDATE_MODE_ADD = 'add'
    ATOMIC_UPDATE_MODE_REMOVE = 'remove'

    DEFAULT_WAIT_TIME_LOCK_FILE = 2

    def __init__(
            self,
            default_obj,
            obj_file_path,
            lock_file_path,
            logger = None,
    ):
        self.default_obj = default_obj
        # fixed and don't change, these file paths
        self.obj_file_path = obj_file_path
        self.lock_file_path = lock_file_path
        self.logger = logger if logger is not None else logging.getLogger()

        self.lockf = LockFile(
            lock_file_path = self.lock_file_path,
            logger = self.logger,
        )
        for f in [self.obj_file_path, self.lock_file_path]:
            self.logger.info('File exists "' + str(f) + '" = ' + str(os.path.exists(f)))

        # Read once from storage
        self.obj = None
        self.obj = self.read_persistent_object()
        self.logger.info(
            'New object persistence created from "' + str(self.obj_file_path)
            + '", lock file "' + str(self.lock_file_path) + '" as: ' + str(self.obj)
        )
        return

    def __assign_default_object_copy(self):
        try:
            self.obj = self.default_obj.copy()
        except Exception as ex_copy:
            errmsg = 'Failed to assign copy of default object: ' + str(ex_copy) \
                     + '. This will potentially modify default object!'
            self.logger.error(errmsg)
            self.obj = self.default_obj

    #
    # Makes sure that read/write happens in one go
    #
    def atomic_update(
            self,
            # Only dict type supported, will add a new items to dict
            new_items,
            # 'add' or 'remove'
            mode,
            max_wait_time_secs = DEFAULT_WAIT_TIME_LOCK_FILE,
    ):
        if not self.lockf.acquire_file_cache_lock(
                max_wait_time_secs = max_wait_time_secs,
        ):
            self.logger.critical(
                'Atomic update could not serialize to "' + str(self.obj_file_path)
                + '", could not obtain lock to "' + str(self.lock_file_path) + '".'
            )
            return False

        try:
            self.obj = self.__deserialize_object_from_file()
            if self.obj is None:
                self.logger.warning(
                    'Atomic update cannot deserialize from file "' + str(self.obj_file_path) + '"'
                )
                self.__assign_default_object_copy()

            self.logger.debug(
                'Cache object type ' + str(type(self.obj))
            )

            if type(self.obj) is dict:
                if type(new_items) is not dict:
                    self.logger.error(
                        'Atomic updates to dict type must be a dict item! Got item type "'
                        + str(type(new_items)) + '": ' + str(new_items)
                    )
                    return False
                for k in new_items.keys():
                    if mode == self.ATOMIC_UPDATE_MODE_ADD:
                        self.obj[k] = new_items[k]
                        self.logger.info('Atomic update added new item ' + str(new_items))
                    elif mode == self.ATOMIC_UPDATE_MODE_REMOVE:
                        if k in self.obj.keys():
                            del self.obj[k]
                            self.logger.info('Atomic update removed item ' + str(new_items))
                    else:
                        self.logger.error('Atomic update invalid mode "'+ str(mode) + '"!')
                        return False
            elif type(self.obj) is list:
                # In this mode, new items is only ONE item, otherwise you get unexpected results
                if mode == ObjectPersistence.ATOMIC_UPDATE_MODE_ADD:
                    self.obj.append(new_items)
                    self.logger.info('Atomic update added new items ' + str(new_items))
                elif mode == ObjectPersistence.ATOMIC_UPDATE_MODE_REMOVE:
                    if new_items in self.obj:
                        self.obj.remove(new_items)
                        self.logger.error(
                            'Atomic updates removed item "' + str(new_items)
                            + str(type(self.obj)) + '"!'
                        )
            else:
                self.logger.error('Atomic updates not supported for cache type "' + str(type(self.obj)) + '"!')
                return False

            res = self.__serialize_object_to_file(obj=self.obj)
            if not res:
                self.logger.error(
                    'Atomic update new item ' + str(new_items)
                    + ' fail, could not serialize update to file "' + str(self.obj_file_path) + '"'
                )
                return False
        except Exception as ex:
            self.logger.error(
                'Atomic update new item ' + str(new_items)
                + ' fail. Exception update to file "' + str(self.obj_file_path) + '": ' + str(ex)
            )
            return False
        finally:
            self.lockf.release_file_cache_lock()
        return True

    #
    # Wrapper write function to applications
    #
    def update_persistent_object(
            self,
            new_obj,
            max_wait_time_secs = DEFAULT_WAIT_TIME_LOCK_FILE
    ):
        self.obj = new_obj
        res = self.serialize_object_to_file(
            obj = self.obj,
            max_wait_time_secs = max_wait_time_secs,
        )
        if not res:
            self.logger.error(
                'Error writing to file "' + str(self.obj_file_path)
                + '", lock file "' + str(self.lock_file_path) + '" for data: ' + str(self.obj)
            )
        return res

    #
    # Wrapper read function for applications
    #
    def read_persistent_object(
            self,
            max_wait_time_secs = DEFAULT_WAIT_TIME_LOCK_FILE,
    ):
        obj_read = self.deserialize_object_from_file(
            max_wait_time_secs = max_wait_time_secs,
        )
        if obj_read is not None:
            self.obj = obj_read
        else:
            self.logger.warning(
                'None object from file "' + str(self.obj_file_path)
                + '", lock file "' + str(self.lock_file_path) + '". Returning memory object.'
            )

        if type(self.default_obj) != type(self.obj):
            self.logger.warning(
                'Object read from file "' + str(self.obj_file_path)
                + '" of type "' + str(type(self.obj)) + ', different with default obj type "'
                + str(type(self.default_obj)) + '". Setting obj back to default obj.'
            )
            self.__assign_default_object_copy()
            self.logger.info('Assigned default copy to object')

            # Need to write back to correct the object type
            self.update_persistent_object(new_obj=self.obj)
            self.logger.info('Updated persistent object with new object ' + str(self.obj))

        return self.obj

    def serialize_object_to_file(
            self,
            obj,
            max_wait_time_secs = DEFAULT_WAIT_TIME_LOCK_FILE,
    ):
        if not self.lockf.acquire_file_cache_lock(
                max_wait_time_secs = max_wait_time_secs
        ):
            self.logger.critical(
                'Could not serialize to "' + str(self.obj_file_path) + '", could not obtain lock to "'
                + str(self.lock_file_path) + '".'
            )
            return False

        try:
            res = self.__serialize_object_to_file(obj=obj)
            return res
        except Exception as ex:
            self.logger.critical(
                'Exception deserializing/loading object from file "'
                + str(self.obj_file_path) + '". Exception message: ' + str(ex) + '.'
            )
            return False
        finally:
            self.lockf.release_file_cache_lock()

    # lock must already be obtained before calling any functions with "__"
    def __serialize_object_to_file(
            self,
            obj,
    ):
        fhandle = open(
            file = self.obj_file_path,
            mode = 'wb'
        )
        pickle.dump(
            obj      = obj,
            file     = fhandle,
            protocol = pickle.HIGHEST_PROTOCOL
        )
        fhandle.close()
        self.logger.debug(
            'Object "' + str(obj)
            + '" serialized successfully to file "' + str(self.obj_file_path) + '"'
        )
        return True

    def deserialize_object_from_file(
            self,
            max_wait_time_secs = DEFAULT_WAIT_TIME_LOCK_FILE,
    ):
        if not self.lockf.acquire_file_cache_lock(
            max_wait_time_secs = max_wait_time_secs,
        ):
            self.logger.critical(
                'Could not deserialize from "' + str(self.obj_file_path) + '", could not obtain lock to "'
                + str(self.lock_file_path) + '".'
            )
            return None

        try:
            return self.__deserialize_object_from_file()
        except Exception as ex:
            self.logger.critical(
                'Exception deserializing/loading object from file "'
                + str(self.obj_file_path) + '". Exception message: ' + str(ex) + '.'
            )
            return None
        finally:
            self.lockf.release_file_cache_lock()

    # lock must already be obtained before calling any functions with "__"
    def __deserialize_object_from_file(
            self,
    ):
        fhandle = open(
            file = self.obj_file_path,
            mode = 'rb',
        )
        obj = pickle.load(
            file = fhandle,
        )
        fhandle.close()
        self.logger.debug(
            'Object "' + str(obj) + '" deserialized successfully from file "' + str(self.obj_file_path)
            + '" to ' + str(obj) + '.'
        )
        return obj


#
# We do extreme testing on ObjectPersistence, by running hundreds of threads updating
# a single file.
# We then check back if there are any errors.
#
class LoadTest:

    DELETED_KEYS_SET = set()

    def __init__(
            self,
            obj_file_path,
            lock_file_path,
            max_wait_time_secs,
            n_threads,
            count_to,
            logger = None,
    ):
        self.obj_file_path = obj_file_path
        self.lock_file_path = lock_file_path
        self.max_wait_time_secs = max_wait_time_secs
        self.n_threads = n_threads
        self.count_to = count_to
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def target(self, thread_no):
        cache = ObjectPersistence(
            default_obj = {},
            obj_file_path = self.obj_file_path,
            lock_file_path = self.lock_file_path,
            logger = self.logger,
        )
        for i in range(self.count_to):
            # To ensure all values are unique, "count_to" is the mathematical base
            value = self.count_to*thread_no + i
            cache.atomic_update(
                new_items = {value: threading.get_ident()},
                mode = ObjectPersistence.ATOMIC_UPDATE_MODE_ADD
            )
            self.logger.info('Value=' + str(value) + ' +++ ' + str(cache.read_persistent_object()))
            # Delete something at random
            if random.choice([0,1]) == 1:
                obj = cache.read_persistent_object()
                key_choices = list(obj.keys())
                if len(key_choices) > 0:
                    random_key_to_delete = random.choice(key_choices)
                    cache.atomic_update(
                        new_items = {random_key_to_delete: obj[random_key_to_delete]},
                        mode = ObjectPersistence.ATOMIC_UPDATE_MODE_REMOVE
                    )
                    LoadTest.DELETED_KEYS_SET.add(random_key_to_delete)
                    self.logger.info('DELETED ' + str(random_key_to_delete))
            time.sleep(random.uniform(0.005,0.010))
        self.logger.info('***** THREAD ' + str(threading.get_ident()) + ' DONE ' + str(self.count_to) + ' COUNTS')

    def test(self):
        threads_list = []
        n_sum = 0
        for i in range(self.n_threads):
            n_sum += self.count_to
            thr = threading.Thread(
                target = self.target,
                args = [i],
            )
            threads_list.append(thr)
            self.logger.info(str(i) + '. New thread "' + str(threads_list[i].getName()) + '" count ' + str(self.count_to))
        expected_values = []
        for i in range(len(threads_list)):
            for j in range(self.count_to):
                expected_values.append(self.count_to*i + j)
            thr = threads_list[i]
            self.logger.info('Starting thread ' + str(i))
            thr.start()

        for thr in threads_list:
            thr.join()

        cache = ObjectPersistence(
            default_obj = {},
            obj_file_path = self.obj_file_path,
            lock_file_path = self.lock_file_path,
            logger = self.logger,
        )
        self.logger.info('********* Final Object File: ' + str(cache.read_persistent_object()))
        values = list(cache.read_persistent_object().keys())
        self.logger.info('Added Keys: ' + str(values))
        self.logger.info('Deleted Keys: ' + str(LoadTest.DELETED_KEYS_SET))
        self.logger.info('Total Added = ' + str(len(values)))
        self.logger.info('Total Deleted = ' + str(len(LoadTest.DELETED_KEYS_SET)))
        values.sort()
        expected_values = list( set(expected_values) - LoadTest.DELETED_KEYS_SET )
        expected_values.sort()
        test_pass = values == expected_values
        self.logger.info('PASS = ' + str(test_pass))
        self.logger.info('Values:   ' + str(values))
        self.logger.info('Expected: ' + str(expected_values))
        return


class UnitTestObjectPersistence:

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def __remove_files(self, obj_file_path, lock_file_path):
        try:
            os.remove(obj_file_path)
        except Exception:
            pass
        try:
            os.remove(lock_file_path)
        except Exception:
            pass

    def test(self):
        test_dir = '.'
        obj_file_path = test_dir + '/loadtest.objpers.obj'
        lock_file_path = test_dir + '/loadtest.objpers.obj.lock'

        #
        # If we don't do this the test will fail as the upcoming threads,
        # many of them will try to delete the file due to timeout, and will
        # accidentally delete one created by another thread
        #
        self.__remove_files(obj_file_path=obj_file_path, lock_file_path=lock_file_path)

        #
        # Test 1: Load Test
        #

        LoadTest(
            obj_file_path  = obj_file_path,
            lock_file_path = lock_file_path,
            count_to = 5,
            n_threads = 10,
            max_wait_time_secs = 10,
            logger = self.logger,
        ).test()

        #
        # Test 2:
        #

        obj_file_path = test_dir + '/objPersTest.b'
        lock_file_path = test_dir + '/lock.objPersTest.b'
        self.__remove_files(obj_file_path=obj_file_path, lock_file_path=lock_file_path)

        obj_pers_1 = ObjectPersistence(
            default_obj = [],
            obj_file_path = obj_file_path,
            lock_file_path = lock_file_path,
            logger = self.logger,
        )

        objects = [
            {
                'a': [1, 2, 3],
                'b': 'test object'
            },
            # Empty objects
            [],
            {},
            88,
            'eighty eight'
        ]

        for obj in objects:
            obj_pers_1.serialize_object_to_file(
                obj = obj,
            )

            b = obj_pers_1.deserialize_object_from_file()
            self.logger.info(str(b))
            ok = obj == b
            assert ok

        obj_file_path = test_dir + '/objPersTestAtomicUpdate.d'
        lock_file_path = test_dir + '/lock.objPersTestAtomicUpdate.d'
        self.__remove_files(obj_file_path=obj_file_path, lock_file_path=lock_file_path)
        obj_pers_2 = ObjectPersistence(
            default_obj    = {},
            obj_file_path  = obj_file_path,
            lock_file_path = lock_file_path
        )
        self.logger.info(
            obj_pers_2.atomic_update(new_items={1: 'hana', 2: 'dul'}, mode=ObjectPersistence.ATOMIC_UPDATE_MODE_ADD)
        )
        ok = obj_pers_2.read_persistent_object() == {1: 'hana', 2: 'dul'}
        assert ok
        self.logger.info(obj_pers_2.atomic_update(new_items={1: 'hana'}, mode=ObjectPersistence.ATOMIC_UPDATE_MODE_REMOVE))
        ok = obj_pers_2.read_persistent_object() == {2: 'dul'}
        assert ok
        self.logger.info(obj_pers_2.atomic_update(new_items={3: 'set'}, mode=ObjectPersistence.ATOMIC_UPDATE_MODE_ADD))
        ok = obj_pers_2.read_persistent_object() == {2: 'dul', 3: 'set'}
        assert ok

        # Purposely write wrong type
        obj_pers_2.update_persistent_object(new_obj=[1,2,3])
        x = ObjectPersistence(
            default_obj    = {},
            obj_file_path  = obj_file_path,
            lock_file_path = lock_file_path
        )
        ok = x.read_persistent_object() == {}
        assert ok
        print('OK TESTS PASSED')
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)

    obj_fp = 'test_tensor'
    lock_fp = 'lock_tensor'
    for f in [obj_fp, lock_fp]:
        # Need to remove, otherwise there will be lingering objects in the file
        os.remove(obj_fp) if os.path.exists(obj_fp) else 1

    obj_pers = ObjectPersistence(
        default_obj = [],
        obj_file_path = obj_fp,
        lock_file_path = lock_fp,
        logger = lgr,
    )
    UnitTestObjectPersistence(logger=lgr).test()

    import numpy as np
    objects = np.random.random(size=(7,3)).tolist()
    [print(o) for o in objects]

    for o in objects:
        obj_pers.atomic_update(
            new_items = o,
            mode = ObjectPersistence.ATOMIC_UPDATE_MODE_ADD,
            max_wait_time_secs = 2,
        )
        print('added item: ' + str(o))

    read_back = obj_pers.deserialize_object_from_file()
    print('Read back:')
    [print(o) for o in read_back]
    assert read_back == objects, 'Read back\n' + str(read_back) + '\nnot same with original\n' + str(objects)
    print('ALL OK')

    exit(0)



import logging
import re
import os
import time
import numpy as np
# pip install mysql-connector-python (NOT mysql-connector)
import mysql.connector
from fitxf import DatastoreInterface, DbParams
from fitxf.math.utils.Lock import Lock
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env


#
# Pull docker image, start it, then create test database
# >> docker pull mysql:8.4
# >> docker run -d --name test-mysql -e MYSQL_ROOT_PASSWORD=password123 -p 3307:3306 mysql:8.4
# >> docker exec -it test-mysql bash
#    bash-4.4# mysql --host=localhost --port=3306 -u root -p
#    mysql> CREATE DATABASE testdb;
# https://www.datacamp.com/tutorial/set-up-and-configure-mysql-in-docker
#
class MySql(DatastoreInterface):

    # Default column name created by ES to store text embedding
    COL_TEXT_ENCODE = 'text_encode'

    # https://www.w3schools.com/mysql/mysql_datatypes.asp
    MYSQL_STRING_DATA_TYPES_WITH_LENGTH = (
        'char', 'varchar', 'blob',
        # 'binary', 'varbinary',
        # 'bit', 'tinyint', 'smallint', 'mediumint', 'int', 'integer', 'bigint',
    )

    def __init__(
            self,
            db_params: DbParams,
            ignore_warnings = False,
            logger = None,
    ):
        super().__init__(
            db_params = db_params,
            ignore_warnings = ignore_warnings,
            logger = logger,
        )
        # Controls locking at least for a single instance
        #self.__mutex_mysql = threading.Lock()
        self.__mutex_name_1 = 'mutex_1'
        self.__mutex_lock = Lock(
            mutex_names = [self.__mutex_name_1],
            logger = self.logger,
        )

        self.create_table_columns_json = {}
        # Automatically convert create table to json format, to be able to extract more info about max lengths, etc.
        self.__convert_create_table_syntax_to_json(
            sql = self.db_params.db_create_table_sql,
        )
        self.logger.info('Create table SQL initialized as: ' + str(self.db_params.db_create_table_sql))
        self.logger.info('Create table SQL to JSON as: ' + str(self.create_table_columns_json))

        self.cursor = None

        # If you have multiple threads/workers accessing the same MySQL, set to True.
        # It seems MySQL can't handle simple synching issues, and need a new reconnection
        # to reflect the latest changes
        # NOTE: Closing cursor after every tx will not work, you need to close the whole connection
        self.close_cursor_after_every_tx = True
        self.close_connections_after_every_tx = True
        return

    def remove_stupid_msql_ticks(self, s):
        return re.sub(pattern="`", repl="", string=s)

    def __check_need_cleanup_every_time(self):
        if self.close_cursor_after_every_tx:
            self.__close_cursor()
        if self.close_connections_after_every_tx:
            self.__close_connection()
        return

    def __execute_sql_with_commit(
            self,
            sql,
            val = None,
            table = 'test',
    ):
        self.reconnect_forever(table=table)
        try:
            self.__mutex_lock.acquire_mutexes(
                id = '__execute_sql_with_commit',
                mutexes = [self.__mutex_name_1],
            )
            # TODO Handle when execute cursor hang forever
            # Put lots of logging here for problematic MySQL that can hang forever when another connection has the lock
            self.logger.info('Now will execute cursor for commit sql "' + str(sql) + '", values: ' + str(val))
            if val is None:
                self.cursor.execute(sql)
            else:
                self.cursor.execute(sql, val)
            self.logger.info('Executed cursor for commit sql "' + str(sql) + '", values ' + str(val))
            self.con.commit()
            self.logger.info('Committed sql "' + str(sql) + '", val ' + str(val))
            self.__check_need_cleanup_every_time()
        except Exception as ex:
            self.logger.error(
                'Error executing sql with commit "' + str(sql) + '", val ' + str(val) + '. Exception: ' + str(ex)
            )
            raise ex
        finally:
            self.__mutex_lock.release_mutexes(mutexes=[self.__mutex_name_1])
        return

    def __execute_sql_data_fetch(
            self,
            sql,
            table = 'test',
    ):
        self.reconnect_forever(table=table)
        try:
            self.__mutex_lock.acquire_mutexes(
                id = '__execute_sql_data_fetch',
                mutexes = [self.__mutex_name_1],
            )
            # TODO Handle when execute cursor hang forever
            # Put lots of logging here for problematic MySQL that can hang forever when another connection has the lock
            self.logger.info('Now will execute cursor for fetch sql "' + str(sql) + '"')
            self.cursor.execute(sql)
            self.logger.info('Executed cursor for commit sql "' + str(sql) + '"')
            resp = self.cursor.fetchall()
            self.logger.info('Executed sql with data fetchall "' + str(sql) + '"')
            self.__check_need_cleanup_every_time()
            # self.logger.info('Response type "' + str(type(resp)) + '": ' + str(resp))
            return resp
        except Exception as ex:
            self.logger.error('Error executing sql data fetch "' + str(sql) + '"')
            raise ex
        finally:
            self.__mutex_lock.release_mutexes(mutexes=[self.__mutex_name_1])

    def __create_table_if_not_exist(
            self,
            table,
    ):
        sql = "SHOW TABLES LIKE '" + str(table) + "'"
        res = self.__execute_sql_data_fetch(table=table, sql=sql)
        self.logger.info('Result for show tables like "' + str(table) + '": ' + str(res))
        exist = len(res) > 0

        if not exist:
            sql = re.sub(pattern="<TABLENAME>", repl=str(table), string=self.db_params.db_create_table_sql)
            self.__execute_sql_with_commit(table=table, sql=sql)
            self.logger.info('Created table "' + str(table) + '"')
        else:
            self.logger.info('Table "' + str(table) + '" already exists')
        return

    def __test_connection(
            self,
            table,
    ):
        try:
            if (self.con is None) or (self.cursor is None):
                return False

            sql = "SHOW TABLES LIKE '" + str(table) + "'"
            self.cursor.execute(sql)
            self.cursor.fetchall()
            return True
        except Exception as ex:
            self.logger.error('Error testing MySQL connection: ' + str(ex))
            return False

    def close(
            self,
    ):
        try:
            self.__mutex_lock.acquire_mutexes(
                id = 'close',
                mutexes = [self.__mutex_name_1],
            )
            self.__close_cursor()
            self.__close_connection()
        finally:
            self.__mutex_lock.release_mutexes(mutexes=[self.__mutex_name_1])

    #
    # Closing cursor and then obtaining a new one will not refresh the data updated by other threads, workers, etc.
    # So in a way this function is useless since we still need to close the whole connection everytime anyway
    #
    def __close_cursor(
            self,
    ):
        if self.cursor is not None:
            try:
                self.cursor.close()
                self.logger.info('Cursor closed successfully')
            except Exception as ex:
                self.logger.error('Error closing cursor: ' + str(ex))

        # Set to None no matter what happens
        self.cursor = None
        return

    def __close_connection(
            self,
    ):
        if self.con is not None:
            try:
                self.con.close()
                self.logger.info('Connection closed successfully')
            except Exception as ex:
                self.logger.error('Error closing connection: ' + str(ex))

        # Set to None no matter what happens
        self.con = None
        return

    def reconnect_forever(
            self,
            table,
            max_tries = 100,
    ):
        try:
            self.__mutex_lock.acquire_mutexes(
                id = 'reconnect_forever',
                mutexes = [self.__mutex_name_1],
            )
            try_count = 0
            while True:
                try:
                    if (self.con is not None) and self.close_cursor_after_every_tx:
                        self.cursor = self.con.cursor()
                except Exception as ex:
                    self.logger.error('Error obtaining cursor: ' + str(ex))
                    # We assume connection already broken
                    self.__close_cursor()
                    self.__close_connection()
                    continue

                if self.__test_connection(table=table):
                    self.logger.debug('MySQL connection still alive')
                    break
                else:
                    # We assume connection already broken
                    self.__close_cursor()
                    self.__close_connection()

                if try_count > max_tries:
                    raise Exception('Max tries to reconnect already reached ' + str(max_tries))

                try:
                    try_count += 1
                    self.__connect()
                    self.logger.info('Reconnected successfully at try #' + str(try_count))
                    break
                except Exception as ex:
                    self.logger.error('Connect try error #' + str(try_count) + ': ' + str(ex))
                    time.sleep(np.random.randint(low=2, high=5))
                    continue
        finally:
            self.__mutex_lock.release_mutexes(mutexes=[self.__mutex_name_1])

    def connect(
            self,
    ):
        try:
            self.__mutex_lock.acquire_mutexes(
                id = 'connect',
                mutexes = [self.__mutex_name_1],
            )
            return self.__connect()
        finally:
            self.__mutex_lock.release_mutexes(mutexes=[self.__mutex_name_1])

    def __connect(
            self,
    ):
        required_conn_params = [
            self.db_params.db_host, self.db_params.db_port, self.db_params.db_username, self.db_params.db_password
        ]
        for _ in required_conn_params:
            assert _ is not None, 'Cannot be empty ' + str(required_conn_params)

        self.logger.info(
            'Try connect to "' + str(self.db_params.db_host) + '", port "' + str(self.db_params.db_port)
            + '", DB "' + str(self.db_params.db_database)
            + '", via username "' + str(self.db_params.db_username) + '", timeout ' + str(self.timeout) + 's.'
        )

        self.con = mysql.connector.connect(
            user = self.db_params.db_username,
            password = self.db_params.db_password,
            host = self.db_params.db_host,
            port = self.db_params.db_port,
            database = self.db_params.db_database,
            connect_timeout = self.timeout,
        )
        # we only need 1 cursor
        self.cursor = self.con.cursor()
        self.logger.info(
            'Connected successfully to "' + str(self.db_params.db_host) + '", port "' + str(self.db_params.db_port)
            + '", DB "' + str(self.db_params.db_database) + '", via username "' + str(self.db_params.db_username)
            + '", timeout ' + str(self.timeout)
            + 's, connection ' + str(id(self.con)) + ', cursor ' + str(id(self.cursor))
        )
        return

    # To prevent burdensome & messy configurations, settings on column restrictions on env vars or config files,
    # we will extract these restrictions programmatically
    # To simplify code, your create table syntax must have ticks for all column names, e.g. `text`, `Faq Name`
    def __convert_create_table_syntax_to_json(
            self,
            sql,
    ):
        self.logger.info('Converting this create table sql: ' + str(sql))

        # remove front part of create table
        # E.g. "CREATE TABLE `<TABLENAME>` (`text` varchar(255) DEFAULT NULL,`label` varchar(255) DEFAULT NULL,)"
        cols_part = re.sub(
            # keep the part inside bracket beginning with a tick "`", thus your create table syntax column names
            # must all have ticks to enclose the column name
            pattern = ".*CREATE TABLE.*[(][\s\t\n\r]*(`.*)",
            repl = "\\1",
            string = sql,
            flags = re.IGNORECASE
        )
        # remove end part of create table
        cols_part = re.sub(pattern="[\s\t\n\r]*[)][\s\t\n\r]*$", repl="", string=cols_part, flags=re.IGNORECASE)
        # E.g. will become "`text` varchar(255) DEFAULT NULL, `label` varchar(255) DEFAULT NULL"
        self.logger.info('Create table sql columns part: ' + str(cols_part))

        cols_part = cols_part.split(sep=",")
        # remove front spaces, new lines, etc.
        cols_part = [str(s).strip() for s in cols_part]
        # E.g. will become ['`text` varchar(255) DEFAULT NULL', '`label_user` varchar(255) DEFAULT NULL']
        self.logger.info('Columns part as list: ' + str(cols_part))

        self.create_table_columns_json = {}
        for i, cpart in enumerate(cols_part):
            if re.match(pattern="^PRIMARY KEY.*", string=cpart):
                self.logger.info('Found constraint "' + str(cpart) + '", breaking from loop...')
            # Check for constraint statement
            if re.match(pattern='^PRIMARY KEY.*', string=cpart):
                self.logger.info('Found constraint "' + str(cpart) + '", breaking from loop..')
                break
            if len(cpart) == 0:
                self.logger.info('Skip empty cpart at #' + str(i))
                continue
            assert cpart[0] == "`", 'Column names must have ticks "' + str(cpart) + '"'
            colname = re.sub(pattern="^`(.*)`.*", repl="\\1", string=cpart)

            col_details = re.sub(pattern="`.*`(.*)", repl="\\1", string=cpart).strip()
            # Split to 2 parts, thus 1st part column type, and 2nd part is the remaining details like DEFAULT values
            col_details = col_details.split(sep=" ", maxsplit=1)
            assert len(col_details) == 2, 'Column details for "' + str(colname) + '" not 2 parts: ' + str(col_details)

            col_type = col_details[0].strip().lower()
            col_other = col_details[1]

            col_type_name = re.sub(pattern="[(].*", repl="", string=col_type)
            if col_type_name in self.MYSQL_STRING_DATA_TYPES_WITH_LENGTH:
                col_type_maxlen = re.sub(pattern=".*[(]([0-9]+)[)].*", repl="\\1", string=col_type)
                col_type_maxlen = int(col_type_maxlen)
            else:
                col_type_maxlen = None

            self.create_table_columns_json[colname] = {
                'type': col_type_name,
                'maxlen': col_type_maxlen,
                'other': col_other,
            }
        self.logger.info(
            'Columns part after convert to json: ' + str(self.create_table_columns_json)
        )
        return

    def get_column_names(
            self,
            tablename,
    ):
        assert tablename is not None, 'Table name cannot be None'
        sql = 'SHOW COLUMNS FROM `' + str(tablename) + '`'
        resp = self.__execute_sql_data_fetch(table=tablename, sql=sql)
        colnames = [self.remove_stupid_msql_ticks(s=a[0]) for a in resp]
        self.logger.info('Raw MySQL query "' + str(sql) + '" response: ' + str(resp))
        self.logger.info('For table "' + str(tablename) + '", column names: ' + str(colnames))
        return colnames

    def get(
            self,
            # e.g. {"answer": "take_seat"} or list of tuples [("answer", "take_seat"), ("question", "...")]
            match_phrase,
            match_condition = {'and': True, 'exact': True},
            tablename = None,
            request_timeout = 20.0,
    ):
        assert tablename is not None, 'Table name cannot be None'
        self.__create_table_if_not_exist(table=tablename)

        if type(match_phrase) is dict:
            match_phrase_tuples = [(k, v) for k, v in match_phrase.items()]
        elif type(match_phrase) in [list, tuple]:
            for pair in match_phrase:
                assert len(pair) == 2, 'Match phrase tuple must be length 2: ' + str(pair)
            match_phrase_tuples = match_phrase
        else:
            raise Exception('Wrong type for match phrase ' + str(type(match_phrase)))

        prm_cond_and = match_condition['and']
        prm_exact_match = match_condition['exact']

        self.logger.debug(
            'Get/Query from index "' + str(tablename) + '", condition AND = ' + str(prm_cond_and)
            + ', exact match = ' + str(prm_exact_match) + ' for query ' + str(match_phrase_tuples)
        )

        colnames = self.get_column_names(tablename=tablename)

        sql = 'SELECT * FROM `' + str(tablename) + '` WHERE '
        for i, (k, v) in enumerate(match_phrase_tuples):
            if i > 0:
                sql += " AND " if prm_cond_and else " OR "
            equal_or_like = '=' if prm_exact_match else 'LIKE'
            sql += '`' + str(k) + '`' + " " + equal_or_like \
                   + " '" + re.sub(pattern="'", repl=r"\'", string=str(v)) + "'"

        self.logger.debug('Try to select from sql "' + str(sql) + '"')
        resp = self.__execute_sql_data_fetch(table=tablename, sql=sql)
        self.logger.debug('Raw MySQL query response: ' + str(resp))
        records = [
            {self.remove_stupid_msql_ticks(s=k): v for k, v in list(zip(colnames, row))}
            for row in resp
        ]
        return records

    def get_all(
            self,
            key = None,
            # <= 0 means get ALL
            max_records = 0,
            tablename = None,
            request_timeout = 20.0,
    ):
        assert tablename is not None, 'Table name cannot be None'
        self.__create_table_if_not_exist(table=tablename)

        colnames = self.get_column_names(tablename=tablename)
        # Add MySQL quotations to columns
        colnames = ["`"+col+"`" for col in colnames]

        sql = 'SELECT ' + ', '.join(colnames) + ' FROM `' + str(tablename) + '`'
        if max_records > 0:
            sql += ' LIMIT ' + str(max_records)
        self.logger.info('Executing SQL: ' + str(sql))
        resp = self.__execute_sql_data_fetch(table=tablename, sql=sql)
        self.logger.debug('Raw MySQL query response: ' + str(resp))
        records = [
            {self.remove_stupid_msql_ticks(s=k): v for k, v in list(zip(colnames, row))}
            for row in resp
        ]
        return records

    def get_indexes(self):
        sql = 'SHOW TABLES'
        resp = self.__execute_sql_data_fetch(sql=sql)
        self.logger.info('Raw MySQL query response: ' + str(resp))
        records = [v[0] for v in resp]
        return records

    def delete_index(
            self,
            tablename,
    ):
        assert tablename is not None, 'Table name cannot be None'
        sql = 'DROP TABLE `' + str(tablename) + '`'
        self.__execute_sql_with_commit(table=tablename, sql=sql)
        self.logger.info('Deleted table "' + str(tablename) + '"')
        return

    def __process_record_values_before_insert_delete(
            self,
            # single record
            record,
    ):
        cols = list(record.keys())
        cols_maxlen = [self.create_table_columns_json[c]['maxlen'] for c in cols]
        self.logger.info('Columns and maxlens: ' + str(list(zip(cols, cols_maxlen))))
        vals = []
        for i, (k, v) in enumerate(record.items()):
            maxlen = cols_maxlen[i]
            if (maxlen is not None) and (type(v) is str):
                v_crop = v[0:min(len(v), maxlen)]
                vals.append(v_crop)
                if v != v_crop:
                    self.logger.warning(
                        'Cropped column value before insert to max length ' + str(maxlen)
                        + ', column "' + str(k) + '", value "' + str(v) + '", after crop "' + str(v_crop) + '"'
                    )
            else:
                vals.append(v)
        return vals

    def add(
            self,
            # list of dicts
            records,
            tablename = None,
            params_other = None,
    ):
        assert tablename is not None, 'Table name cannot be None'
        self.__create_table_if_not_exist(table=tablename)

        assert len(records) > 0, 'No records to insert into: ' + str(records)
        for record in records:
            cols = list(record.keys())
            for c in cols:
                assert c in self.create_table_columns_json.keys(), \
                    'Column to be inserted does not exist in json "' + str(c) \
                    + '" ' + str(self.create_table_columns_json)
            cols_maxlen = [self.create_table_columns_json[c]['maxlen'] for c in cols]
            self.logger.info('Columns and maxlens for sql INSERT: ' + str(list(zip(cols, cols_maxlen))))
            # Add MySQL quotations to columns
            cols = ["`"+cl+"`" for cl in cols]
            sql = 'INSERT INTO `' + str(tablename) + '` (' + ', '.join(cols) + ') VALUES (' \
                  + ', '.join(['%s']*len(cols)) + ')'
            val = self.__process_record_values_before_insert_delete(record=record)
            # val = [v for k,v in record.items()]
            self.logger.info('Try to insert with sql "' + str(sql) + '", values ' + str(val))
            res = self.__execute_sql_with_commit(table=tablename, sql=sql, val=val)
            self.logger.info('Insert result for sql "' + str(sql) + '", values ' + str(val) + ': ' + str(res))
        return

    def delete(
            self,
            match_phrase,
            match_condition = {'and': True, 'exact': True},
            tablename = None,
    ):
        assert type(match_phrase) is dict, 'Match phrase wrong type "' + str(type(match_phrase)) + '"'
        assert tablename is not None, 'Table name cannot be None'
        self.__create_table_if_not_exist(table=tablename)

        sql = 'DELETE FROM `' + str(tablename) + '` WHERE '
        values = self.__process_record_values_before_insert_delete(record=match_phrase)
        for i, (k, v) in enumerate(match_phrase.items()):
            if i > 0:
                and_or = 'AND' if match_condition['and'] else 'OR'
                sql += " " + str(and_or) + " "
            # sql += str(k) + " = '" + re.sub(pattern="'", repl="\\'", string=str(v)) + "'"
            # Use placeholder "%s" so we don't need to do messy escape of special characters
            eq_or_like = '=' if match_condition['exact'] else 'LIKE'
            sql += '`' + str(k) + '` ' + str(eq_or_like) + ' %s'
            # values.append(str(v))
        self.logger.info('Try to delete from sql "' + str(sql) + '"')
        self.__execute_sql_with_commit(table=tablename, sql=sql, val=values)
        # TODO How to get how many deleted successfully?
        return {'deleted': -1}

    def delete_by_raw_query(
            self,
            raw_query,
            tablename = None,
            params_other = None,
    ):
        assert tablename is not None, 'Table name cannot be None'
        self.__create_table_if_not_exist(table=tablename)

        params_other = {} if params_other is None else params_other
        self.__execute_sql_with_commit(
            sql = raw_query,
            table = tablename,
            val = None,
        )
        # TODO How to get how many deleted successfully?
        return {'deleted': -1}

    def delete_all(
            self,
            key,
            tablename = None,
    ):
        raise Exception('TODO')


class MySqlUnitTest:
    def __init__(
            self,
            db_params: DbParams,
            db_table_name,
            reconnect_after_every_tx,
            logger = None,
    ):
        self.db_params = db_params
        self.db_table_name = db_table_name
        self.reconnect_after_every_tx = reconnect_after_every_tx
        self.logger = logger if logger is not None else logging.getLogger()

        self.db_params.db_create_table_sql = """CREATE TABLE `<TABLENAME>` (
                `tex t` varchar(10) DEFAULT NULL,
                `label_user` varchar(255) DEFAULT NULL,
                `label_standardized` int DEFAULT NULL,
                `text_encoding` varchar(5000) DEFAULT NULL
                )"""
        self.my1 = MySql(
            db_params = self.db_params,
            ignore_warnings = True,
            logger = self.logger,
        )
        self.my1.close_connections_after_every_tx = self.reconnect_after_every_tx
        self.my2 = MySql(
            db_params = db_params,
            ignore_warnings = True,
            logger = self.logger,
        )
        self.my2.close_connections_after_every_tx = self.reconnect_after_every_tx
        return

    def connect(self, my: MySql):
        my.connect()
        self.logger.info('Connection ' + str(my) + ' connected')

    def test(self):
        self.connect(my=self.my1)
        self.connect(my=self.my2)

        records_1 = [
            {'tex t': "don't use ' character", 'label_user': 'test', 'label_standardized': 1, 'text_encoding': '+++'},
            {'tex t': "Hi hi", 'label_user': 'hi', 'label_standardized': 2, 'text_encoding': '+++'},
            {'tex t': "Yo yo", 'label_user': 'hi', 'label_standardized': 2, 'text_encoding': '+++'},
            {'tex t': "white", 'label_user': 'color', 'label_standardized': 3, 'text_encoding': '+++'},
            {'tex t': "white", 'label_user': 'color', 'label_standardized': 3, 'text_encoding': '+++'},
            {'tex t': "black", 'label_user': 'color', 'label_standardized': 3, 'text_encoding': '+++'},
        ]
        records_2 = [
            {'tex t': "missing 1", 'label_user': 'missing', 'label_standardized': 4, 'text_encoding': '+++'},
            {'tex t': "missing 2", 'label_user': 'missing', 'label_standardized': 4, 'text_encoding': '+++'},
        ]
        # delete all records
        for rec in records_1 + records_2:
            self.my1.delete(match_phrase=rec, tablename=self.db_table_name)
        for i, my in enumerate([self.my1, self.my2]):
            rows = my.get_all(tablename=self.db_table_name, max_records=0)
            assert len(rows) == 0, 'Rows not empty for i=' + str(i) + ': ' + str(rows)

        # for 1st connection, add only 1st batch
        self.my1.add(records=records_1, tablename=self.db_table_name)
        # self.my1.delete(
        #     match_phrase = {'text': "don't use ' character", 'label_standardized': 1},
        #     tablename = self.db_table_name,
        # )

        # for 2nd connection, only add 2nd batch
        self.my1.add(records=records_2, tablename=self.db_table_name)

        #--------------------------------------------------------------------------------------------------
        # IMPORTANT IN PRODUCTION
        #   An open connection will not be updated with latest data if another connection does an INSERT,
        #   DELETE, etc.
        #   This means in production, you must set the flag close_connections_after_every_tx to True
        #--------------------------------------------------------------------------------------------------
        # Must reconnect for both connections to be in sync, otherwise 2nd connection will return empty DB

        # Now both connections must retrieve the same data
        rows_retrieved = [[], []]
        for i, my_x in enumerate([self.my1, self.my2]):
            rows = my_x.get_all(tablename=self.db_table_name, max_records=0)
            print('Connection #' + str(i + 1) + ': (ALL) Type ' + str(type(rows)) + ':')
            [print('   ' + str(i + 1) + '. ' + str(r)) for i, r in enumerate(rows)]
            rows_retrieved[i] = rows
        if self.reconnect_after_every_tx:
            assert str(rows_retrieved[0]) == str(rows_retrieved[1]), \
                'Rows different\n' + '\n'.join([str(line) for line in rows_retrieved[0]]) \
                + '\nwith\n' + '\n'.join([str(line) for line in rows_retrieved[1]])
        else:
            assert len(rows_retrieved[1]) == 0, \
                'Expected no data in 2nd MySql connection due to not sync: ' + str(rows_retrieved[1])

        for mp, match_cond, exp_rows_count in [
            ({'label_user': 'color', 'tex t': "white"}, {"and": True, "exact": True}, 2),
            ({'label_user': 'color', 'tex t': "white"}, {"and": False, "exact": True}, 3),
            ({'tex t': "missing%"}, {"and": False, "exact": False}, 2),
            ({'tex t': "missing%"}, {"and": False, "exact": True}, 0),
            # Use tuples in match phrase for similar columns
            ([('tex t', 'white'), ('tex t', 'black')], {"and": True, "exact": True}, 0),
            ([('tex t', 'white'), ('tex t', 'black')], {"and": False, "exact": True}, 3),
            ([('tex t', 'white'), ('tex t', 'black')], {"and": False, "exact": False}, 3),
            ([('tex t', 'wh%'), ('tex t', 'bla%')], {"and": False, "exact": False}, 3),
        ]:
            rows = self.my1.get(
                match_phrase = mp,
                match_condition = match_cond,
                tablename = self.db_table_name,
            )
            assert len(rows) == exp_rows_count, \
                'For mp ' + str(mp) + ', prms ' + str(match_cond) + '.Got total rows ' + str(len(rows)) \
                + ' but expected ' + str(exp_rows_count) + ', rows ' + str(rows)

        limit = 3
        rows = self.my1.get_all(tablename=self.db_table_name, max_records=limit)
        print('(LIMIT) Type ' + str(type(rows)) + ':')
        [print('   ' + str(i + 1) + '. ' + str(r)) for i, r in enumerate(rows)]
        assert len(rows) == 3, 'Expected rows limit ' + str(limit) + ' but got ' + str(len(rows)) + ': ' + str(rows)

        rows = self.my1.get(match_phrase={'label_user': 'test'}, tablename=self.db_table_name)
        print('Type ' + str(type(rows)) + ': ' + str(rows))
        # assert len(rows) == len(records_1+records_2)

        rows = self.my1.get_all(tablename=self.db_table_name)
        # [print(rw) for rw in rows]
        self.my1.delete_by_raw_query(
            raw_query = "DELETE FROM `" + str(self.db_table_name) + "` WHERE `tex t` LIKE '%miss%'",
            tablename = self.db_table_name,
        )
        before_delete_count = len(rows)
        # print('***** AFTER DELETE')
        rows = self.my1.get_all(tablename=self.db_table_name)
        # [print(rw) for rw in rows]
        assert len(rows) == before_delete_count - 2, \
            'After delete row count ' + str(len(rows)) + ', before delete row count ' + str(before_delete_count)

        # Close 2nd connection so it won't lock table to be deleted
        self.my2.close()
        self.my1.delete_index(tablename=self.db_table_name)
        self.my1.close()
        self.logger.info('ALL TESTS PASSED OK')
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut.mysql')
    table_name = 'en-marktest'

    for recon in [True, False]:
        MySqlUnitTest(
            logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False),
            db_params = DbParams(
                identifier = 'testmysql',
                db_type = 'mysql',
                db_host = os.environ["DB_HOST"],
                db_port = os.environ["DB_PORT"],
                db_username = os.environ["DB_USERNAME"],
                db_password = os.environ["DB_PASSWORD"],
                db_database = os.environ["DB_DATABASE"],
            ),
            db_table_name = table_name,
            reconnect_after_every_tx = recon,
        ).test()

    # res = my1.get_indexes()
    # delete_unit_test_indexes = False
    # if delete_unit_test_indexes:
    #     # Must close 2nd connection first
    #     my2.close()
    #     for tbl in res:
    #         if re.match(pattern='intent_unit_test.*', string=tbl):
    #             print('Matched "' + str(tbl) + '"')
    #             my1.delete_index(tablename=tbl)
    #     exit(0)

    exit(0)



import logging
import threading
import os
import time
from datetime import datetime
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


#
# Instantiate this class instead of your desired class, then call get_singleton()
#
class Singleton:

    # Default global singleton store, if not provided by user
    SINGLETON_STORE_BY_CLASSTYPE = {}
    SINGLETON_LAST_UPDATE_BY_CLASSTYPE = {}

    # A gloabl mutex means you cannot call get singleton inside a get singleton of a same class,
    # means singleton inside singleton of same class not allowed
    MUTEX_SINGLETON_STORE_BY_CLASSTYPE = {}

    def __init__(
            self,
            class_type,
            logger = None,
    ):
        self.class_type = class_type
        self.logger = logger if logger is not None else logging.getLogger()

        # No need to care about race condition here, at most the latter lock object will just overwrite the previous
        # Worst case is that 2 singletons are stored on 2 different storage duee to race condition,
        # of which the earlier one will become irrelevant later
        if not self.class_type in self.SINGLETON_STORE_BY_CLASSTYPE.keys():
            self.SINGLETON_STORE_BY_CLASSTYPE[self.class_type] = {}
            self.SINGLETON_LAST_UPDATE_BY_CLASSTYPE[self.class_type] = {}
            self.logger.info('Created new store for class type "' + str(self.class_type) + '"')

        # No need to care about race condition here, at most the latter lock object will just overwrite the previous
        if not self.class_type in self.MUTEX_SINGLETON_STORE_BY_CLASSTYPE.keys():
            self.MUTEX_SINGLETON_STORE_BY_CLASSTYPE[self.class_type] = threading.Lock()
            self.logger.info('Created new mutex for class type "' + str(self.class_type) + '"')

        self.singleton_store = self.SINGLETON_STORE_BY_CLASSTYPE[self.class_type]
        self.last_update_dict = self.SINGLETON_LAST_UPDATE_BY_CLASSTYPE[self.class_type]
        return

    # def get_derived_key(self, user_key_id):
    #     return str(self.class_type) + '___' + str(user_key_id)

    def get_singleton(
            self,
            # user provided key, if need multiple singletons for the same class.
            # user's responsibility to make sure no naming clashes between IDs of different class types
            key_id,
            *args,
    ):
        try:
            self.logger.info('Trying to acquire singleton mutex for class type ' + str(self.class_type))
            # Use different mutex, so that a singleton of one class can instantiate a singleton of a different class
            self.MUTEX_SINGLETON_STORE_BY_CLASSTYPE[self.class_type].acquire()
            return self.__get_singleton_thread_safe(
                key_id, *args
            )
        finally:
            self.MUTEX_SINGLETON_STORE_BY_CLASSTYPE[self.class_type].release()
            self.logger.info('Singleton mutex released for class type ' + str(self.class_type))

    def __get_singleton_thread_safe(
            self,
            # The reason we need a user provided ID is that sometimes for a given class type,
            # the user may want to return different instance if the __init__() parameters are
            # different. And we let user have that control.
            # user's responsibility to make sure no naming clashes between IDs of different class types
            key_id,
            *args,
    ):
        # Update last accessed time
        self.last_update_dict[key_id] = datetime.now()
        self.logger.info(
            'Last active for singleton "' + str(key_id) + '" set to ' + str(self.last_update_dict[key_id])
        )

        # derived_key = self.get_derived_key(user_key_id=key_id)
        if key_id in self.singleton_store.keys():
            self.logger.info(
                'Returning already available singleton class class type "' + str(self.class_type)
                + '", key id "' + str(key_id) + '": ' + str(self.singleton_store[key_id])
            )
            type_existing = type(self.singleton_store[key_id])
            assert type_existing == self.class_type, \
                'Wrong type for singleton "' + str(type_existing) + '" with "' + str(type(self.class_type)) + '"'
            return self.singleton_store[key_id]
        else:
            self.singleton_store[key_id] = self.class_type(*args)
            self.logger.info(
                'Created new singleton class for class type "' + str(self.class_type)
                + '", key id "' + str(key_id) + '": ' + str(self.singleton_store[key_id])
            )
            return self.singleton_store[key_id]


class SingletonMemoryMgmt:

    def __init__(
            self,
            class_type,
            delete_secs_inactive: float,
            thread_check_every_n_secs:float = 10.,
            logger = None,
    ):
        self.class_type = class_type
        self.delete_secs_inactive = delete_secs_inactive
        self.thread_check_every_n_secs = thread_check_every_n_secs
        self.signal_stop_thread = False
        self.logger = logger if logger is not None else logging.getLogger()

        if self.class_type not in Singleton.SINGLETON_STORE_BY_CLASSTYPE.keys():
            Singleton.SINGLETON_STORE_BY_CLASSTYPE[self.class_type] = {}
            Singleton.SINGLETON_LAST_UPDATE_BY_CLASSTYPE[self.class_type] = {}
            Singleton.MUTEX_SINGLETON_STORE_BY_CLASSTYPE[self.class_type] = threading.Lock()

        self.__store = Singleton.SINGLETON_STORE_BY_CLASSTYPE[self.class_type]
        self.__last_update = Singleton.SINGLETON_LAST_UPDATE_BY_CLASSTYPE[self.class_type]
        self.__mutex = Singleton.MUTEX_SINGLETON_STORE_BY_CLASSTYPE[self.class_type]

        self.memory_mgmt_records = {}
        self.thread = threading.Thread(target=self.run_memory_mgmt_thread)
        self.thread.start()
        self.logger.info('Started memory management thread for Singletons.')
        return

    def update_memory_mgmt_params(
            self,
            delete_secs_inactive,
    ):
        self.delete_secs_inactive = delete_secs_inactive
        self.logger.info('Set delete secs inactive to new value ' + str(self.delete_secs_inactive) + 's')
        return

    def __delete_key_id(self, key_id):
        del self.__store[key_id]
        del self.__last_update[key_id]
        # force garbage collection
        gc.collect()
        # Mutex can't be deleted
        # del self.__mutex[key_id]
        self.logger.info(
            'Deleted key id "' + str(key_id) + '", remaining keys: ' + str(self.__store.keys())
        )
        return

    def clear_all(self):
        try:
            self.__mutex.acquire()
            for key_id in self.__store.keys():
                self.__delete_key_id(key_id=key_id)
        finally:
            self.__mutex.release()

    def run_memory_mgmt_thread(self):
        while True:
            if self.signal_stop_thread:
                self.logger.info(
                    'Memory mgmt for "' + str(self.class_type) + '" stop signal received, ending thread..'
                )
                break
            time.sleep(self.thread_check_every_n_secs)
            try:
                self.__mutex.acquire()
                keys_to_delete = []
                for key_id in self.__store.keys():
                    self.logger.debug('Checking key id "' + str(key_id) + '"')
                    last_active = self.__last_update[key_id]
                    diff = datetime.now() - last_active
                    secs_non_active = round(diff.days * 86400 + diff.seconds + diff.microseconds / 1000000, 4)
                    if secs_non_active > self.delete_secs_inactive:
                        self.logger.info(
                            'Delete inactive for ' + str(secs_non_active) + 's, key id "' + str(key_id)
                            + '", set object for deletion..'
                        )
                        keys_to_delete.append(key_id)
                    else:
                        self.logger.debug(
                            'Keep inactive for ' + str(secs_non_active) + 's, key id "' + str(key_id) + '"'
                        )
                for key_id in keys_to_delete:
                    self.__delete_key_id(key_id=key_id)
            except Exception as ex:
                self.logger.error('Error in memory mgmt thread ' + str(ex))
            finally:
                self.__mutex.release()

    def stop_memory_mgmt(self):
        self.signal_stop_thread = True
        return


class SingletonUnitTest:
    class SampleTestClass:
        def __init__(self, a, b, throw_exception=False, logger=None):
            self.a = a
            self.b = b
            self.logger = logger if logger is not None else logging.getLogger()
            self.address = id(self)
            self.logger.info('Initialized new object ' + str(self) + ' at ' + str(hex(self.address)))
            if throw_exception:
                raise Exception('Flag throw exception set')
            return

    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        # start memory mgmt
        self.deleted_after_secs = 10000.
        self.test_class_type = SingletonUnitTest.SampleTestClass

        self.mm = SingletonMemoryMgmt(
            class_type = self.test_class_type,
            delete_secs_inactive = self.deleted_after_secs,
            thread_check_every_n_secs = 0.2,
            logger = self.logger,
        )
        return

    def test(self):
        store = Singleton.SINGLETON_STORE_BY_CLASSTYPE[self.test_class_type]
        sgt = Singleton(
            class_type = self.test_class_type,
            logger = self.logger,
        )

        for i in range(10):
            user_key = str(i % 3)
            throw_exc = True if user_key == '2' else False

            got_exception = False
            try:
                obj = sgt.get_singleton(user_key, i, i + 1, throw_exc, self.logger)
            except Exception as ex:
                self.logger.error('Error get singleton ' + str(ex))
                obj = None
                got_exception = True

            assert throw_exc == got_exception, \
                'Requested to throw exception = ' + str(throw_exc) + ' but got ' + str(got_exception)
            assert len(store) == len(set(store.values())), 'All in store must be unique: ' + str(store)
            self.logger.info('All in store at #' + str(i) + ': ' + str(store))

            # key = sgt.get_derived_key(user_key_id=user_key)
            if not throw_exc:
                assert id(obj) == id(store[user_key]), \
                    'Object key "' + str(user_key) + '" not same ' + str(id(obj)) + ' vs ' + str(id(store[user_key]))
            else:
                assert user_key not in store.keys(), 'User key "' + str(user_key) + '" should not exist'

        sleep_total = 0.
        id_0_before_deletion = id(store['0'])
        while sleep_total < 3.0:
            obj = sgt.get_singleton('0', 0, 1, False, self.logger)
            assert id_0_before_deletion == id(obj), \
                'Must be same cached object ' + str(id(obj)) + ' as old ' + str(id_0_before_deletion)
            time.sleep(1.0)
            sleep_total += 1.

        # Make memory mgmt delete the objects
        self.mm.update_memory_mgmt_params(delete_secs_inactive=1.)

        # After sleep, memory should have been cleaned up
        time.sleep(1.)
        assert len(store) == 0, 'Store should be cleaned up of memory but got ' + str(store)
        self.mm.stop_memory_mgmt()

        #
        # Test that new object will be created since we already deleted the above
        #
        obj = sgt.get_singleton('0', 0, 1, False, self.logger)
        assert id_0_before_deletion != id(obj), \
            'Must be new object created already, but got same ' + str(id(obj)) + ' as old ' + str(id_0_before_deletion)

        print('SINGLETON TESTS PASSED OK')


if __name__ == '__main__':
    e = Env()
    Env.set_env_vars_from_file(env_filepath=e.REPO_DIR + '/.env.fitxf.math.ut')
    lgr = Logging.get_logger_from_env_var()
    SingletonUnitTest(logger=lgr).test()
    exit(0)



import logging
import sys
if sys.version_info[0] + sys.version_info[1]/10 >= 3.8:
    import importlib.metadata as mtdata
else:
    import pkg_resources


# Helper class to manage environments & versions
class PkgVersion:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logging.getLogger() if logger is None else logger
        self.python_ge_3_8 = sys.version_info[0] + sys.version_info[1] / 10 >= 3.8
        self.logger.info('Is python version >= 3.8 = "' + str(self.python_ge_3_8) + '"')
        return

    def get_pkg_version_string(self, pkg_name):
        pkg_ver = mtdata.version(pkg_name) if self.python_ge_3_8 else pkg_resources.get_distribution(pkg_name).version
        return pkg_ver

    def convert_version_to_list(
            self,
            version: str,
            maxlen: int = 0,
    ):
        assert maxlen <= 3
        v_list = version.split(sep='.', maxsplit=3)
        if maxlen > 0:
            [v_list.append('') for _ in range(3-len(v_list))]
            v_list = v_list[0:maxlen]
        return v_list

    def check_package_version(
            self,
            pkg_name: str,
            version: str,
            return_version = False,
    ):
        v_cmp = self.convert_version_to_list(version=version)
        pkg_ver = mtdata.version(pkg_name) if self.python_ge_3_8 else pkg_resources.get_distribution(pkg_name).version
        v_cur = self.convert_version_to_list(
            version = pkg_ver,
            # Compare only based on given version, means if actual version is "7.17.9",
            # given version "7", we only compare major "7" against "7"
            maxlen  = len(v_cmp),
        )
        self.logger.debug('Pkg "' + str(pkg_name) + '" version ' + str(v_cur))
        verdict = tuple(v_cmp)==tuple(v_cur)
        return (verdict, v_cur) if return_version else verdict

    def check_python(
            self,
            version: str,
            return_version = False,
    ):
        v_cmp = self.convert_version_to_list(version=version)
        v_cur = self.convert_version_to_list(
            version = '.'.join([str(sys.version_info.major), str(sys.version_info.minor), str(sys.version_info.micro)]),
            maxlen  = len(v_cmp),
        )
        str(sys.version_info.major), str(sys.version_info.minor), str(sys.version_info.micro)
        self.logger.debug('Python current version: ' + str(v_cur))
        verdict = tuple(v_cmp)==tuple(v_cur)
        return (verdict, v_cur) if return_version else verdict


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    pkg = 'numpy'
    print(PkgVersion().check_python(version='3', return_version=True))
    print(PkgVersion().check_python(version='3.8', return_version=True))
    print(PkgVersion().check_python(version='3.10', return_version=True))
    print(PkgVersion().check_python(version='3.8.16', return_version=True))

    print(PkgVersion().check_package_version(pkg_name=pkg, version='1.24', return_version=True))
    print(PkgVersion().check_package_version(pkg_name=pkg, version='1.22', return_version=True))
    print(PkgVersion().check_package_version(pkg_name=pkg, version='1.24.0', return_version=True))
    exit(0)



import numpy as np
import torch
import warnings
import logging


class TensorUtils:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        # ignore divide by 0 warnings
        warnings.filterwarnings('ignore')
        return

    def convert_to_numpy(self, x):
        assert type(x) in [np.ndarray, torch.Tensor], 'Wrong type "' + str(type(x)) + '": ' + str(x)
        return x.cpu().detach().numpy() if type(x) is torch.Tensor else x

    """
    Normalize Tensor of any dimension
    """
    def normalize(
            self,
            x,
            return_tensors = 'np',
    ):
        x_tmp = self.convert_to_numpy(x=x)

        # normalize the last row dimension
        norm = np.linalg.norm(x_tmp, axis=-1, keepdims=True)
        x_norm = x_tmp / norm

        # handle 0's
        x_norm[np.isnan(x_norm)] = 0.
        return torch.from_numpy(x_norm) if return_tensors=='pt' else x_norm

    def similarity_cosine(self, x, ref, return_tensors='np'):
        return self.dot_sim(x=x, ref=ref, return_tensors=return_tensors)

    # arccos similarity
    def dot_sim(
            self,
            x,
            ref,
            return_tensors = 'np',
    ):
        x_np = self.convert_to_numpy(x=x)
        ref_np = self.convert_to_numpy(x=ref)
        assert x_np.ndim == 2, 'Dim must be 2 but got ' + str(x_np.ndim) + ': ' + str(x_np)
        assert ref_np.ndim == 2, 'Dim must be 2 but got ' + str(ref_np.ndim) + ': ' + str(ref_np)
        x_norm = self.normalize(
            x = x_np,
            return_tensors = 'np',
        )
        ref_norm = self.normalize(
            x = ref_np,
            return_tensors = 'np',
        )
        self.logger.debug('t shape ' + str(ref_norm.shape) + ', txt_lm shape ' + str(x_norm.shape))
        # can be negative which means in the opposite direction from reference
        m_dot = np.matmul(x_norm, np.transpose(ref_norm))
        # Avoid nan for arccos for values 1+smallvalue or -1-smallvalue
        # values returned mapped from cosine [-1,1] range, which corresponds ranges from [0, pi],
        # never negative unlike arcsin()
        # m_dot = np.arccos(np.minimum(np.maximum(m_dot, -1.), 1.))
        # sort descending, flip the argsort which by default is ascending
        result_ordered = np.flip( np.argsort(m_dot, axis=-1), axis=-1 )
        m_dot_ordered = np.array([
            [m_dot[i,result_ordered[i,j]] for j in range(m_dot.shape[1])] for i in range(m_dot.shape[0])
        ])

        return (torch.from_numpy(result_ordered.copy()), torch.from_numpy(m_dot_ordered.copy())) \
            if return_tensors == 'pt' else (result_ordered, m_dot_ordered)

    def similarity_distance(
            self,
            x,
            ref,
            return_tensors = 'np',
    ):
        x_np = self.convert_to_numpy(x=x)
        ref_np = self.convert_to_numpy(x=ref)
        assert x_np.ndim == 2, 'Dim must be 2 but got ' + str(x_np.ndim) + ': ' + str(x_np)
        assert ref_np.ndim == 2, 'Dim must be 2 but got ' + str(ref_np.ndim) + ': ' + str(ref_np)

        # Make compatible for taking difference
        # Convert for example row vectors
        #    [
        #      [1,2,3],
        #      [4,5,6],
        #    ]
        # into a repeated sequence (same count as how many rows in reference tensor)
        #    [
        #      [ [1,2,3], [1,2,3], [1,2,3] ],
        #      [ [4,5,6], [4,5,6], [4,5,6] ],
        #    ]
        self.logger.debug('x:\n' + str(x_np))
        self.logger.debug('ref:\n' + str(ref_np))
        x_np_rp = np.repeat(x_np, len(ref_np), axis=0)
        x_np_rp = np.reshape(x_np_rp, (len(x_np), len(ref_np), x_np.shape[-1]))
        self.logger.debug('x repeated:\n' + str(x_np_rp))
        m_difsq = (x_np_rp - ref_np) ** 2
        self.logger.debug('difference squared:\n' + str(m_difsq))
        m_dist = np.sum(m_difsq, axis=-1) ** 0.5
        self.logger.debug('distances:\n' + str(m_dist))
        result_ordered = np.argsort(m_dist, axis=-1)
        self.logger.debug('index ordered:\n' + str(result_ordered))
        m_dist_ordered = np.array([
            [m_dist[i,result_ordered[i,j]] for j in range(m_dist.shape[1])] for i in range(m_dist.shape[0])
        ])
        self.logger.debug('distance ordered:\n' + str(m_dist_ordered))

        return (torch.from_numpy(result_ordered.copy()), torch.from_numpy(m_dist_ordered.copy())) \
            if return_tensors == 'pt' else (result_ordered, m_dist_ordered)


class TensorUtilsUnitTest:

    def _helper_test_numpy_arrays(self, x1, x2):
        assert x1.shape == x2.shape
        dif = np.sum((x1 - x2)**2)
        assert dif < 0.0000000001

    def test_norm(self):
        m = TensorUtils()

        v_np_1d = np.array([1, 2, 3, 4])
        expected = np.array([0.18257419, 0.36514837, 0.54772256, 0.73029674])
        nm = m.normalize(x=v_np_1d, return_tensors='np')
        self._helper_test_numpy_arrays(x1=nm, x2=expected)

        v_np = np.array([[1, 2, 3], [5, 4, 6], [-1, -2, -3], [0, 0, 0]])
        v_pt = torch.FloatTensor([[1, 2, 3], [5, 4, 6], [-1, -2, -3], [0, 0, 0]])
        expected = np.array([
            [ 0.26726124,  0.53452248,  0.80178373],
            [ 0.56980288,  0.45584231,  0.68376346],
            [-0.26726124, -0.53452248, -0.80178373],
            [ 0.,          0.,          0.,        ],
        ])
        nm = m.normalize(x=v_np, return_tensors='np')
        self._helper_test_numpy_arrays(x1=nm, x2=expected)
        nm = m.normalize(x=v_pt, return_tensors='pt')
        self._helper_test_numpy_arrays(x1=nm.cpu().detach().numpy(), x2=expected)

        v_np_3d = np.array([
            [[1, 2, 3], [5, 4, 6], [-1, -2, -3], [0, 0, 0]],
            [[1, 1, 1], [2, 2, 2], [-1, -1, -1], [0, 0, 0]],
        ])
        expected = np.array([
            [
                [ 0.26726124,  0.53452248,  0.80178373],
                [ 0.56980288,  0.45584231,  0.68376346],
                [-0.26726124, -0.53452248, -0.80178373],
                [ 0.,          0.,          0.,        ],
            ],
            [
                [ 0.57735027,  0.57735027,  0.57735027],
                [ 0.57735027,  0.57735027,  0.57735027],
                [-0.57735027, -0.57735027, -0.57735027],
                [ 0.,          0.,          0.,        ],
            ],
        ])
        nm = m.normalize(v_np_3d)
        self._helper_test_numpy_arrays(x1=nm, x2=expected)
        print('ALL TESTS PASSED OK (norm)')
        return

    def test_similarity_cosine_and_similarity_distance(self):
        mu = TensorUtils()
        x = np.array([
            [1, 2, 3],
            [6, 5, 4],
        ])
        ref = np.array([
            [1.2, 1.9, 3.3],
            [6.7, 4.8, 4.1],
            [-6.7, 4.8, -4.1],
        ])
        res, m = mu.dot_sim(
            x = x,
            ref = ref,
            return_tensors = 'np',
        )
        expected_res = np.array([[0, 1, 2], [1, 0, 2]])
        expected_m = np.array([[0.99742005,  0.83034349, -0.2729101], [0.99780448,  0.85345704, -0.40357849]])
        self._helper_test_numpy_arrays(x1=res, x2=expected_res)
        self._helper_test_numpy_arrays(x1=m, x2=expected_m)

        res, m = mu.similarity_distance(
            x = x,
            ref = ref,
            return_tensors = 'np',
        )
        expected_res = np.array([[0, 1, 2], [1, 0, 2]])
        expected_m = np.array([[0.37416574,  6.44515322, 10.8415866], [0.73484692,  5.75673519, 15.06452787]])
        self._helper_test_numpy_arrays(x1=res, x2=expected_res)
        self._helper_test_numpy_arrays(x1=m, x2=expected_m)
        print('ALL TESTS PASSED OK (similarity cosine & distance)')
        return


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    TensorUtilsUnitTest().test_norm()
    TensorUtilsUnitTest().test_similarity_cosine_and_similarity_distance()
    exit(0)



import logging


class ModelEncoderInterface:
    def __init__(
            self,
            # can be LLM model files, or whatever
            cache_folder: str = None,
            logger = None,
    ):
        self.cache_folder = cache_folder
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def get_model_name(self):
        raise Exception('Must be implemented by derived class')

    def get_model_path(self):
        raise Exception('Must be implemented by derived class')

    def encode(self, content_list, return_tensors: str = 'np'):
        raise Exception('Must be implemented by derived class')



import logging
import os
import json
import pickle
import numpy as np
from fitxf.math.fit.transform.FitXformInterface import FitXformInterface
from fitxf.math.fit.utils.FitUtils import FitUtils
from fitxf.math.fit.cluster.Cluster import Cluster
from fitxf.math.fit.cluster.ClusterCosine import ClusterCosine
from fitxf.math.utils.Lock import Lock
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env


class FitXformCluster(FitXformInterface):

    KEY_CLUSTERNO_TO_USERLABELS_INFO = 'clusterno_to_userlabels_info'

    def __init__(
            self,
            logger = None,
    ):
        super().__init__(
            logger = logger,
        )

        self.__mutex_model = 'model'
        self.__lock = Lock(
            mutex_names = [self.__mutex_model],
            logger = self.logger,
        )
        self.fit_utils = FitUtils(logger=self.logger)
        self.cluster = Cluster(logger=self.logger)

        # Model parameters
        self.model_params_ready = False
        self.cluster_labels = None
        self.cluster_inertia = None
        self.cluster_inertia_per_point = None
        self.centers_median_distance = None
        return

    def is_model_ready(self):
        return self.model_params_ready

    def estimate_min_max_clusters(
            self,
            n,
    ):
        max_clusters = 3*int(np.log(n))
        min_clusters = max(2, int(np.log(n)))
        self.logger.info(
            'Min/max clusters estimated as ' + str(min_clusters) + ' and ' + str(max_clusters) + ', n=' + str(n)
        )
        return min_clusters, max_clusters

    def fit_optimal(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            target_grid_density = 2,
            # allowed values 'median', 'mean', 'min'
            measure = 'median',
            # Model dependent interpretation, or ignore if not relevant for specific model
            min_components = 2,
            max_components = 100,
            test_mode = False,
    ) -> dict:
        try:
            self.__lock.acquire_mutexes(
                id = 'fit_optimal',
                mutexes = [self.__mutex_model],
            )

            # Interpret as min/max clusters
            min_clusters, max_clusters = min_components, max_components

            return self.__fit_optimal(
                X = X,
                X_labels = X_labels,
                X_full_records = X_full_records,
                target_grid_density = target_grid_density,
                min_clusters = min_clusters,
                max_clusters = max_clusters,
                test_mode = test_mode,
            )
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def fit(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            # Model dependent interpretation, or ignore if not relevant for specific model
            # For example, can mean how many clusters, or how many PCA components, or how many to sample
            # in a discrete Fourier transform, etc.
            n_components = 2,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        n_centers = n_components
        if X_labels is None:
            X_labels = list(range(len(X)))

        self.logger.info(
            'Start kmeans optimal with X shape ' + str(X.shape) + ', n clusters ' + str(n_centers)
            + ', test mode ' + str(test_mode)
        )
        desired_cluster = self.cluster.kmeans(
            x = X,
            n_centers = n_centers,
            x_labels = X_labels,
            start_centers = start_centers,
            km_iters = 100,
            test_mode = test_mode,
        )
        self.logger.info('Desired cluster of requested n=' + str(n_centers) + ': ' + str(desired_cluster))
        return self.__record_cluster(
            desired_cluster = desired_cluster,
            X = X,
            X_labels = X_labels,
            X_full_records = X_full_records,
        )

    # Will dynamically look for optimal number of pca components, on the condition of the target grid density
    def __fit_optimal(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            target_grid_density = 2,
            # Model dependent interpretation, or ignore if not relevant for specific model
            min_clusters = 2,
            max_clusters = 100,
            km_iters = 100,
            # by default if 25% of the clusters are single point clusters, we quit
            thr_single_clusters = 0.25,
            plot = False,
            # pass through mode
            test_mode = False,
    ) -> dict:
        assert target_grid_density > 0, 'Target grid density not valid ' + str(target_grid_density)
        self.logger.info(
            'Start kmeans optimal with X shape ' + str(X.shape) + ', min clusters ' + str(min_clusters)
            + ', max clusters ' + str(max_clusters) + ', using cluster class "' + str(self.cluster.__class__) + '"'
        )
        res = self.cluster.kmeans_optimal(
            x = X,
            x_labels = X_labels,
            km_iters = km_iters,
            min_clusters = min_clusters,
            max_clusters = max_clusters,
            thr_single_clusters = thr_single_clusters,
            plot = plot,
            test_mode = test_mode,
        )
        desired_cluster = res[0]
        self.logger.info(
            'Desired optimal cluster using cluster class "' + str(self.cluster.__class__) + '": ' + str(desired_cluster)
        )
        return self.__record_cluster(
            desired_cluster = desired_cluster,
            X = X,
            X_labels = X_labels,
            X_full_records = X_full_records,
        )

    def fine_tune(
            self,
            X: np.ndarray,
            X_labels: list = None,
            X_full_records: list = None,
            n_components: int = None,
            # pass through mode
            test_mode = False,
    ) -> dict:
        if self.model_n_components_or_centers is not None:
            n_add = n_components - self.model_n_components_or_centers
            if n_add > 0:
                centers_additional = np.random.rand(n_add, self.model_centers.shape[-1])
                start_centers = np.append(self.model_centers, centers_additional, axis=0)
            elif n_add == 0:
                start_centers = self.model_centers
            else:
                start_centers = self.model_centers[:n_add]
            start_centers_shape = start_centers.shape
        else:
            # set as nothing to add to additional model (which does not exist)
            n_add = 0
            # if None means will train from scratch
            start_centers = None
            start_centers_shape = None

        self.logger.info(
            'Start fine tuning by additional ' + str(n_add) + ' centers, shape of start centers '
            + str(start_centers_shape) + '.'
        )
        res = self.fit(
            X = X,
            X_labels = X_labels,
            X_full_records = X_full_records,
            n_components = n_components,
            start_centers = start_centers,
            test_mode = test_mode,
        )
        return res

    def __record_cluster(
            self,
            X: np.ndarray,
            desired_cluster: dict,
            X_labels,
            X_full_records,
    ):
        # Copy over original data, don't retain large tensors in memory like X
        # self.X = np.array(X)
        self.X_labels = X_labels
        self.X_full_records = X_full_records
        self.model_train_total_iterations = desired_cluster.get('total_iterations', None)
        self.model_centers = desired_cluster['cluster_centers']
        # not applicable to cluster model
        self.model_principal_components = np.array([])
        self.cluster_labels = desired_cluster['cluster_labels']
        self.model_n_components_or_centers = desired_cluster['n_centers']
        self.model_centroid = np.mean(self.model_centers)
        self.cluster_no_map_to_userlabel = desired_cluster["cluster_label_to_original_labels"]

        self.cluster_inertia = desired_cluster['points_inertia']
        self.cluster_inertia_per_point = self.cluster_inertia / len(X_labels)
        self.distance_error = self.cluster_inertia_per_point
        self.distance_error_mean = np.mean(self.distance_error)

        self.centers_median_distance = desired_cluster['centers_median']
        self.logger.info(
            'Inertia ' + str(self.cluster_inertia) + ', inertia per point ' + str(self.cluster_inertia_per_point)
            + ', centers median distance ' + str(self.centers_median_distance)
        )

        # Same as cluster labels
        # self.X_transform = self.__transform(X=X)   # also can use transform
        self.X_transform = np.array(self.cluster_labels, dtype=np.int64)
        # assert self.X_transform.tolist() == self.cluster_labels, \
        #     'Inconsistent ' + str(list(zip(self.X_transform.tolist(), self.cluster_labels.tolist())))

        self.logger.info('X transform: ' + str(self.X_transform))
        self.X_inverse_transform = self.__inverse_transform(
            x_transform = self.X_transform,
        )

        X_lengths = np.sum((X * X), axis=-1) ** 0.5
        X_inverse_lengths = np.sum((self.X_inverse_transform * self.X_inverse_transform), axis=-1) ** 0.5
        self.angle_error = np.sum(X * self.X_inverse_transform, axis=-1) / (X_lengths * X_inverse_lengths)
        self.angle_error_mean = np.mean(self.angle_error)

        # not applicable to cluster model
        self.X_grid_vectors = np.array([])
        self.X_grid_numbers = np.array([])

        self.model_params_ready = True

        return self.model_to_b64json(
            numpy_to_base64_str = False,
            dump_to_b64json_str = False,
        )

    # Inverse transform is just the cluster center
    def inverse_transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        try:
            self.__lock.acquire_mutexes(
                id = 'inverse_transform',
                mutexes = [self.__mutex_model],
            )
            return self.__inverse_transform(
                x_transform = X,
            )
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def __inverse_transform(
            self,
            x_transform: np.ndarray,
    ) -> np.ndarray:
        # Transform is just the cluster label (or cluster center index)
        # Make sure is int type to index into model centers
        x_transform_inttype = x_transform.astype(dtype=np.int64)
        x_estimated = self.model_centers[x_transform_inttype]
        return x_estimated

    # Get PCA values of arbitrary points
    def transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        try:
            self.__lock.acquire_mutexes(
                id = 'transform',
                mutexes = [self.__mutex_model],
            )
            return self.__transform(X=X)
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def __transform(
            self,
            X: np.ndarray,
    ) -> np.ndarray:
        pred_labels, pred_probs = self.predict_standard(
            X = X,
            ref_X = self.model_centers,
            ref_labels = list(range(len(self.model_centers))),
            top_k = 3,
        )
        # Cluster transform is just the cluster label
        return np.array([r[0] for r in pred_labels])

    def __calc_grid(
            self,
            X: np.ndarray,
    ):
        raise Exception('TODO')

    def predict(
            self,
            X: np.ndarray,
            # can be cluster numbers to zoom into
            X_search_local_space: np.ndarray = None,
            labels_search_local_space: list = None,
            top_k = 5,
            return_full_record = False,
            use_grid = False,
    ):
        assert return_full_record == False, 'Full record return not supported for cluster prediction'
        try:
            self.__lock.acquire_mutexes(
                id = 'predict',
                mutexes = [self.__mutex_model],
            )

            if X_search_local_space is None:
                # use model centers if local space not specified
                X_ref = self.model_centers
                # labels are just the indexes of the centers - cluster labels
                labels_ref = list(range(len(self.model_centers)))
                using_cluster_labels = True
            else:
                X_ref = X_search_local_space
                # user provided labels
                labels_ref = labels_search_local_space
                using_cluster_labels = False
                assert labels_ref is not None
                assert len(X_ref) == len(labels_ref), \
                    'Inconsistent lengths, X_search_local_space shape ' + str(X_search_local_space.shape) \
                    + ', labels_search_local_space length ' + str(len(labels_search_local_space))

            pred_labels, pred_probs = self.predict_standard(
                X = X,
                ref_X = X_ref,
                ref_labels = labels_ref,
                ref_full_records = self.X_full_records,
                top_k = top_k,
                return_full_record = return_full_record,
            )
            self.logger.debug('Predicted clusters for top k ' + str(top_k) + ': ' + str(pred_labels))

            if using_cluster_labels:
                assert self.cluster_no_map_to_userlabel is not None
                # Include user label info
                pred_label_and_clusterno = []
                # map from cluster numbers to actual user labels
                self.logger.debug(
                    'Try map cluster numbers ' + str(pred_labels) + ' to actual user labels using map: '
                    + str(self.cluster_no_map_to_userlabel)
                )
                for cluster_numbers in pred_labels:
                    user_labels_cno = []
                    for cno in cluster_numbers:
                        # each cluster number is mapped to user labels probabilities, e.g.
                        #   {'food': 1.0, 'genetics': 0.0, 'medicine': 0.0, 'sports': 0.0, 'tech': 0.0}
                        tmp = self.cluster_no_map_to_userlabel[cno]
                        # we just take the top one, e.g. "food"
                        top_label_tmp = list(tmp.keys())[0]
                        user_labels_cno.append({'cluster_label': cno, 'user_label_estimate': top_label_tmp})
                    pred_label_and_clusterno.append(user_labels_cno)
                self.logger.info(
                    'Converted to user labels: ' + str(pred_label_and_clusterno)
                    + ' from cluster numbers ' + str(pred_labels)
                )

                # Do a 2nd zoom search
                # condition_same_cluster = self.X_transform == top_cluster
                # To simplify matters, we leave it to the caller to pull required data and call this
                # function again passing in X_search_local_space & labels_search_local_space
            else:
                pred_label_and_clusterno = pred_labels
                # pred_labels_user = pred_labels

            # Cluster transform is just the cluster label
            # return np.array([r[0] for r in pred_labels])
            return pred_label_and_clusterno, pred_probs
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_model])

    def model_to_b64json(
            self,
            numpy_to_base64_str = False,
            dump_to_b64json_str = False,
    ):
        base_model_dict = super().model_to_b64json(
            numpy_to_base64_str = numpy_to_base64_str,
            dump_to_b64json_str = False,
        )
        # Add more info
        # Keep dictionary as string, because json cannot dump or load np.int32, and that problem is hard
        # to trace where the int32 comes from, so keep the whole dict as string and load back later by literal_eval
        base_model_dict[self.KEY_CLUSTERNO_TO_USERLABELS_INFO] = self.base64.encode(
            b = pickle.dumps(self.cluster_no_map_to_userlabel),
        )
        if dump_to_b64json_str:
            # retain UTF-8 if any
            return self.base64.encode(b=json.dumps(base_model_dict, ensure_ascii=False).encode(encoding='utf-8'))
        else:
            return base_model_dict

    def load_model_from_b64json(
            self,
            model_b64json,
    ) -> dict:
        model_dict = super().load_model_from_b64json(
            model_b64json = model_b64json,
        )
        self.logger.info('Loaded model dict from base class as: ' + str(model_dict))
        [self.logger.info('Key "' + str(k) + '": Value ' + str(v)) for k, v in model_dict.items()]

        # Load back dict using literal_eval, to avoid int32 json problems
        self.cluster_labels = list(self.X_transform)
        self.cluster_no_map_to_userlabel = pickle.loads(
            self.base64.decode(s=model_dict[self.KEY_CLUSTERNO_TO_USERLABELS_INFO]),
        )
        model_dict[self.KEY_CLUSTERNO_TO_USERLABELS_INFO] = self.cluster_no_map_to_userlabel
        self.logger.info(
            'Loaded model from saved base64-json string, for cluster class "' + str(self.cluster.__class__)
            + ' data length ' + str(len(self.X_transform))
            + ', n components/centers ' + str(self.model_n_components_or_centers)
            + ', centers ' + str(len(self.model_centers))
            + ', model base64-json string: ' + str(model_b64json)
        )
        return model_dict


class FitXformClusterCosine(FitXformCluster):

    def __init__(
            self,
            logger = None,
    ):
        super().__init__(
            logger = logger,
        )
        # Overwrite to use ClusterCosine
        self.cluster = ClusterCosine(logger=self.logger)
        return


if __name__ == '__main__':
    from fitxf.math.lang.encode.LangModelPt import LangModelPt as LmPt
    texts = [
        "Let's have coffee", "Free for a drink?", "How about Starbucks?",
        "I am busy", "Go away", "Don't disturb me",
        "Monetary policies", "Interest rates", "Deposit rates",
    ]
    labels = ['coffee', 'coffee', 'coffee', 'busy', 'busy', 'busy', 'ir', 'ir', 'ir']
    lmo = LmPt(cache_folder=Env().MODELS_PRETRAINED_DIR)

    embeddings = lmo.encode(content_list=texts, return_tensors='np')

    fitter = FitXformCluster(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False))
    res = fitter.fit_optimal(
        X = embeddings,
        X_labels = labels,
        test_mode = True,
    )
    # Save model to base64-json string
    model_save_str = fitter.model_to_b64json(numpy_to_base64_str=True, dump_to_b64json_str=True)

    # Load from saved model
    fitter2 = FitXformCluster(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False))
    fitter2.load_model_from_b64json(model_b64json=model_save_str)
    embeddings_clusterno = fitter2.cluster_labels
    print('Embeddings cluster labels ' + str(embeddings_clusterno))

    print('Cluster info:')
    [print(k,v) for k,v in res.items()]
    pred_cnos, pred_probs = fitter2.predict(
        X  = lmo.encode(content_list=['latte with cream']),
        use_grid = False
    )
    top_cluster = pred_cnos[0][0]['cluster_label']
    print('Search among centers: ', top_cluster, pred_cnos, pred_probs)

    # Zoom search in local space
    cond = [True if cno==top_cluster else False for cno in embeddings_clusterno]
    print(cond)
    print('Search among cluster "' + str(top_cluster) + '":', fitter2.predict(
        X = lmo.encode(content_list=['latte with cream']),
        X_search_local_space = embeddings[np.array(cond)],
        labels_search_local_space = [l for i,l in enumerate(labels) if cond[i]==True],
        use_grid = False
    ))
    exit(0)

    fitter.create_scatter_plot2d(
        x_transform = fitter.X_inverse_transform,
        labels_list = texts,
        show = True,
        # add some noise to separate the points a little more
        add_noise = True,
    )

    x = np.array([[1,2,3], [3,2,1], [-1,-2,-2], [-3,-4,-2]])
    res = fitter.fit(X=x, X_labels=['+', '+', '-', '-'], n_components=2)
    print('Cluster info:')
    [print(k,v) for k,v in res.items()]
    print('Fit arbitrary:', fitter.predict(X=np.array([[9,9,8], [-55,-33,-55]]), use_grid=False))
    print('Fit arbitrary:', fitter.predict(X=np.array([[9,9,8], [-55,-33,-55]]), use_grid=True))

    exit(0)
